import random
import zipfile
from collections import defaultdict
from dataclasses import dataclass, field
from functools import lru_cache
from io import BytesIO
from pathlib import Path
from typing import Dict, Iterable, List, Literal, Optional, Set, Tuple, Union

import nibabel as nib
import numpy as np
import pydicom

from spine_segmentation.instance_separation.instance_separation import separate_rois_with_labels
from spine_segmentation.resources.preloaded import get_measure_statistics

MetadataType = Literal["age", "sex", "weight", "size"]
MetadataTypes = Union[MetadataType, Iterable[MetadataType]]


@dataclass
class GTFormat:
    """The format of the ground truth.

    include: Which ground truth to include in the segmentation (wirbelkörper, bandscheiben).
    order: How to order the classes (any, y-sort), with y-sort being where the classes are ordered by
           their y-coordinate from up to down.
    separation: How to separate the classes (binary, semantic, instance), with instance each
                instance of a class is a separate class.
    type: The type of the ground truth (segmentation, classification, regression).

    Note that any other type than "segmentation" implies instance separation as well as y-sort
    """

    include: Literal["wk", "bs", "bs+wk"] = "bs+wk"
    order: Literal["any", "y-sort"] = "any"
    separation: Literal["binary", "semantic", "instance"] = "semantic"
    type: Literal["segmentation", "classification", "single-classification", "regression"] = "segmentation"
    s1_class: Optional[int] = None


def _normalize_image(img, target_shape, is_gt=False, return_nifti=False, return_full=False, crop_height_to_slice=None):
    if target_shape[0] is None:
        target_shape = (img.shape[0], *target_shape[1:])

    # Pad until target shape
    padding = np.subtract(target_shape, img.shape)
    padding[padding < 0] = 0
    img = np.pad(img, list(zip([0, 0, 0], padding)), mode="constant")
    assert img.shape >= target_shape, f"{img.shape=} {target_shape=}"

    # Reduce until target shape
    cx, cy, _ = [size // 2 for size in img.shape]
    sx, sy, sz = [s // 2 for s in target_shape]
    oddx, oddy, oddz = [s % 2 for s in target_shape]
    img = img[cx - sx : cx + sx + oddx, cy - sy : cy + sy + oddy, : sz * 2 + oddz]
    assert img.shape == target_shape, f"{img.shape=} {target_shape=}"

    # Normalize size
    if is_gt:
        img = (img != 0).astype(int)
        assert max(np.unique(img)) == img.max(), f"{np.unique(img)=} {img.max()=}"
    else:
        img = img.astype(np.float32)
        img -= img.min()
        img /= img.max()
        assert img.min() == 0 and img.max() == 1, f"{img.min()=} {img.max()=}"

    # print(img.shape, crop_height_to_slice)
    full_img = img
    if crop_height_to_slice is not None:
        img = img[:, :, crop_height_to_slice]
    # print(img.shape)

    if return_full:
        return img, full_img

    return img


def _load_and_normalize(path, *args, return_nifti=False, **kwargs):
    nifti = nib.load(path).get_fdata()
    img = _normalize_image(nifti, *args, **kwargs)
    if return_nifti:
        return img, nifti
    return img


class MissingGTError(Exception): ...


@lru_cache
def get_pid_to_index_lookup():
    stats = get_measure_statistics()
    ids = stats.dicom_metadata["DICOM;PatientID"]
    return {pid.split("_")[0]: i for i, pid in enumerate(ids)}


class MetadataSample:
    def __init__(self, path, target_shape, gt_type: MetadataTypes):
        """
        Creates a new ImageGT object from the given path.

        image has shape (18, 320, 896) and values in [0, 1].

        gt has shape (18, 320, 896) and values in {0, 1, 2}.
        gt_onehot has shape (3, 18, 320, 896) and values in {0, 1, 2}.

        If each_roi_is_separate_class is True, then the values will be in {0, 1, 2, ..., #num_rois},
        and the first dimension of gt_onehot will be #num_rois.
        """
        self.target_shape = target_shape

        self.pid = path.name
        self.image, self.nifti = _load_and_normalize(path / f"{self.pid}.nii.gz", target_shape, return_nifti=True)
        # self.wk = _load_and_normalize(path / f"{pid}_WK.nii.gz", is_gt=True)
        # self.bs = _load_and_normalize(path / f"{pid}_BS.nii.gz", is_gt=True)
        self.gt = self._compose_groundtruth(gt_type)

    def _compose_groundtruth(self, gt_types: MetadataTypes):
        import torch

        stats = get_measure_statistics()
        lookup = get_pid_to_index_lookup()
        try:
            row = stats.dicom_metadata.iloc[lookup[self.pid]]
        except KeyError:
            raise MissingGTError(f"Missing ground truth for this sample {self.pid=}")
        gt = []

        def handle_gt(gt_type):
            value = row[f"DICOM;Patient{gt_type.capitalize()}"]
            if gt_type == "sex":
                assert value in "MF"
                gt.append((value == "M"))
            elif gt_type == "age":
                assert value.endswith("Y")
                gt.append(int(value[:-1]) / 100)
            elif gt_type == "weight":
                gt.append(float(value) / 100)
            elif gt_type == "size":
                gt.append(float(value))
            else:
                raise ValueError(f"Unknown gt_type {gt_type=} {value=}")

        if isinstance(gt_types, str):
            gt_types = [gt_types]
        for type_ in sorted(set(gt_types)):
            handle_gt(type_)

        return torch.tensor(gt, dtype=torch.float32)


def get_random_crop_slice(crop_height_to_px, full_height, start_at=None):
    if start_at is None:
        crop_start_at = random.randint(0, full_height - crop_height_to_px)
    else:
        crop_start_at = start_at
    return slice(crop_start_at, crop_start_at + crop_height_to_px)


class Sample:
    def __init__(
        self,
        path,
        target_shape,
        gt_format: Optional[GTFormat],
        crop_height_to_px=None,
        zip_to_gt_path_lookup: Dict[Path, Path] = None,
    ):
        """
        Creates a new ImageGT object from the given path.

        image has shape (18, 320, 896) and values in [0.0, 1.0].
        gt has shape (18, 320, 896) and values in {0, 1, 2}.
        """
        self.target_shape = target_shape
        self.gt_format = gt_format
        self.path = Path(path).expanduser()
        self.dicom = None
        self.zip_to_gt_path_lookup = zip_to_gt_path_lookup

        if crop_height_to_px is None:
            crop_height_to_px = self.target_shape[2]
        self.crop_slice = get_random_crop_slice(crop_height_to_px, self.target_shape[2])

        init_methods = [self._init_if_nifti(self.path), self._init_if_zip(self.path)]
        assert sum(init_methods) == 1, f"Ensure exactly one loading init method was used, not {sum(init_methods)=})"

    def _init_if_zip(self, path) -> bool:
        if path.suffix != ".zip":
            return False

        series_id_to_dicoms: Dict[str, List[pydicom.FileDataset]] = defaultdict(list)
        with zipfile.ZipFile(path, "r") as zip_ref:
            for filename in zip_ref.namelist():
                with zip_ref.open(filename) as file:
                    dicom = pydicom.dcmread(BytesIO(file.read()))
                    series_id_to_dicoms[dicom.SeriesInstanceUID].append(dicom)

        images = []
        for series_id, dicoms in sorted(series_id_to_dicoms.items())[:1]:
            self.dicom = dicoms[0]
            for dicom in sorted(dicoms, key=lambda d: d.InstanceNumber):
                images.append(dicom.pixel_array.T[::-1, ::-1])

        self.image, self.full_image = _normalize_image(
            np.stack(images, axis=0), self.target_shape, return_full=True, crop_height_to_slice=self.crop_slice
        )
        if self.zip_to_gt_path_lookup:
            self.npz = np.load(self.zip_to_gt_path_lookup[path], allow_pickle=True)
            seg = self.npz["segmentation"]

            self.full_wk = (seg == 1).astype(int)
            self.wk = self.full_wk[:, :, self.crop_slice]

            self.full_bs = (seg == 2).astype(int)
            self.bs = self.full_bs[:, :, self.crop_slice]

            self.gt, self.full_gt = self._compose_groundtruth()
        else:
            self.full_gt = np.zeros_like(self.full_image)
            self.gt = self.full_gt[:, :, self.crop_slice]
        return True

    def _init_if_nifti(self, path) -> bool:
        pid = path.name

        image_path = path / f"{pid}.nii.gz"
        if not image_path.exists():
            return False

        self.image, self.full_image = _load_and_normalize(
            image_path, self.target_shape, return_full=True, crop_height_to_slice=self.crop_slice
        )
        if self.gt_format is not None:
            self.wk, self.full_wk = _load_and_normalize(
                path / f"{pid}_WK.nii.gz",
                self.target_shape,
                is_gt=True,
                return_full=True,
                crop_height_to_slice=self.crop_slice,
            )
            p = path / f"{pid}_BS.nii.gz"
            if p.exists():
                self.bs, self.full_bs = _load_and_normalize(
                    p, self.target_shape, is_gt=True, return_full=True, crop_height_to_slice=self.crop_slice
                )
            else:
                self.bs = self.full_bs = None
            self.gt, self.full_gt = self._compose_groundtruth()
        return True

    def _compose_groundtruth(self):
        compose_gt_func = getattr(self, f"_compose_gt_{self.gt_format.type}", None)
        if compose_gt_func is not None:
            return compose_gt_func()
        else:
            raise ValueError(f"Unknown gt_format.type '{self.gt_format.type=}'")

    def _compose_gt_classification(self):
        gtf = self.gt_format
        assert gtf.separation == "instance", f"{gtf.separation=}"
        assert gtf.order == "y-sort", f"{gtf.order=}"
        seg_gt, full_seg_gt = self._compose_gt_segmentation()
        # gt_length = 24 * ("bs" in gtf.include) + 25 * ("wk" in gtf.include)
        # print(seg_gt.shape)
        # print(seg_gt)
        uniq = np.unique(seg_gt)
        gt = np.array([n in uniq for n in range(1, 50)]).astype(float)
        # classes = [n for n in uniq if 1 <= n <= 49]
        # gt[classes] = 1
        ret = np.repeat(gt[None], self.target_shape[0], axis=0)
        return ret, ret

    def _compose_gt_segmentation(self):
        gt = np.zeros_like(self.full_wk)
        if self.full_bs is not None:
            wk, bs = self.full_wk, self.full_bs
        else:
            wk, bs = (self.full_wk % 2 == 0) & (self.full_wk > 0), self.full_wk % 2 == 1

        gtf = self.gt_format

        if gtf.separation == "instance":
            if hasattr(self, "npz") and "instances" in self.npz:
                instances = self.npz["instances"]
                id_to_label = self.npz["id_to_label"].item()
            else:
                wkbs = wk - ((wk != 0) & (bs != 0)) + bs * 2
                assert wkbs.shape[2] == 896, f"{wkbs.shape=}"
                instances, id_to_label = separate_rois_with_labels(wkbs)
            curr_index = 1
            for id_, label in sorted(id_to_label.items()):
                is_bs = "-" in label
                if not is_bs and "wk" in gtf.include:
                    if label == "S1" and gtf.s1_class is not None:
                        gt[instances == id_] = gtf.s1_class
                    else:
                        gt[instances == id_] = curr_index
                        curr_index += 1
                if is_bs and "bs" in gtf.include:
                    gt[instances == id_] = curr_index
                    curr_index += 1

        if gtf.separation == "semantic":
            if "wk" in gtf.include:
                gt[wk != 0] = 1
            if "bs" in gtf.include:
                gt[bs != 0] = gt.max() + 1

        if gtf.separation == "binary":
            if "wk" in gtf.include:
                gt[wk != 0] = 1
            if "bs" in gtf.include:
                gt[bs != 0] = 1

        return gt[:, :, self.crop_slice], gt

    def _onehot_encode(self):
        gt = self.gt
        gt_onehot = np.stack([(gt == value).astype(int) for value in sorted(np.unique(gt))])
        assert gt_onehot.min() == 0 and gt_onehot.max() == 1, f"{gt_onehot.min()=} {gt_onehot.max()=}"
        assert len(np.unique(gt_onehot)) == 2, f"{np.unique(gt_onehot)=}"
        return gt_onehot


@dataclass
class SampleIterator(Iterable[Tuple[np.ndarray, np.ndarray]]):
    data_dirs: List[Path]
    crop_height_to_px: int = None
    target_shape: Tuple[Optional[int]] = (None, 320, 896)
    add_adjacent_slices: bool = False
    already_processed: Set[Path] = field(default_factory=set)
    skip_first_percent: float = 0

    def __iter__(self):
        start_at = int(len(self.data_dirs) * self.skip_first_percent)
        for path in self.data_dirs[start_at:]:
            if path in self.already_processed:
                continue
            try:
                sample = Sample(path, self.target_shape, gt_format=GTFormat(), crop_height_to_px=self.crop_height_to_px)
                batch_img = []
                batch_gt = []
                for i in range(sample.image.shape[0]):
                    if self.add_adjacent_slices:
                        image, gt = sample.image[max(0, i - 1) : i + 2], sample.gt[i : i + 1]

                        if i == 0:
                            image = np.concatenate([image[:1], image], axis=0)
                        if i == sample.image.shape[0] - 1:
                            image = np.concatenate([image, image[-1:]], axis=0)

                        assert image.shape[0] == 3, f"{image.shape=}"

                    else:
                        image = sample.image[i : i + 1]
                        gt = sample.gt[i : i + 1]
                    batch_img.append(image)
                    batch_gt.append(gt)
                yield np.stack(batch_img, axis=0), np.stack(batch_gt, axis=0), sample, path
            except KeyboardInterrupt:
                raise
            except:
                print("Error")
                import traceback

                traceback.print_exc()


def main():
    gt_format = GTFormat()

    from dataclasses import fields
    from typing import Literal, get_args

    for field in fields(GTFormat):
        print(f"Field name: {field.name}")

        # If the field type is a Literal, iterate over its options
        if getattr(field.type, "__origin__", None) is Literal:
            literals = get_args(field.type)
            for literal in literals:
                print(f"Literal: {literal}")

        # Print the value of the field for the instance
        print(f"Value: {getattr(gt_format, field.name)}")


if __name__ == "__main__":
    main()



import importlib

import torch


def get_object_from_string(full_path):
    """
    Given a string of format 'module.submodule.function_name',
    imports the module and returns the function object.
    """
    module_path, function_name = full_path.rsplit(".", 1)
    module = importlib.import_module(module_path)
    function = getattr(module, function_name)
    return function


class ProxyClass(torch.nn.Module):
    def __init__(self, cls, *args, **kwargs):
        super().__init__()
        func = get_object_from_string(cls)
        self._cls = func(*args, **kwargs)
        self.__class__.__name__ = type(self._cls).__name__

    def __getattr__(self, name):
        if name == "_cls":
            return super().__getattr__(name)
        return getattr(self._cls, name)

    def __call__(self, *args, **kwargs):
        return self._cls(*args, **kwargs)

    def __getitem__(self, item):
        return self._cls[item]



import numpy as np


class Plane:
    def __init__(self, centroid: np.ndarray, normal_vector: np.ndarray, *, name: str = None, index: int = None):
        assert centroid.shape == (3,)
        assert normal_vector.shape == (3,)
        assert abs(np.linalg.norm(normal_vector)) > 1e-5
        self.centroid = centroid
        self.normal_vector = normal_vector / np.linalg.norm(normal_vector)

        if name is not None:
            self.name = name

        if index is not None:
            self.index = index

    @staticmethod
    def from_centroid_and_normal(centroid, normal_vector):
        return Plane(centroid, normal_vector)

    def as_corners(self, size=10) -> np.ndarray:
        v = self.normal_vector
        if np.linalg.norm(v - np.array([1, 0, 0])) > 1e-5:
            w = np.array([1, 0, 0])
        else:
            w = np.array([0, 1, 0])

        u = np.cross(v, w)
        u_prime = np.cross(v, u)

        vectors = np.array([u, u_prime, -u, -u_prime])
        return vectors * size + self.centroid

    @property
    def points(self):
        return self.as_corners()

    def are_points_above_plane(self, points):
        """
        >>> c = np.array([5, 5, 5])
        >>> n = np.array([1, 0, 0])
        >>> plane = Plane(c, n)
        >>> plane.are_points_above_plane(np.array([[0, 0, 1], [0, 0, 0], [0, 0, -1]]))
        array([ True, False, False])

        >>> import matplotlib.pyplot as plt
        >>> np.random.seed(0)
        >>> arr = np.random.randint(0, 2, size=(10, 10, 10)).astype(bool)
        >>> ax = plt.figure().add_subplot(projection="3d")
        >>> ticks = range(0, 12, 2)
        >>> ax.set_xticks(ticks)
        >>> ax.set_yticks(ticks)
        >>> ax.set_zticks(ticks)
        >>> is_above = plane.are_points_above_plane(np.array(np.nonzero(arr)).T)
        # >>> print(is_above)
        >>> colors = np.empty(arr.shape, dtype=object)
        >>> colors[np.nonzero(arr)] = ["#ff0000" if yes else "#0000ff" for yes in is_above]
        >>> ax.voxels(arr, facecolors=colors)
        {}
        >>> xx, yy = np.meshgrid(range(0, 11), range(0, 11))
        >>> zz = (n[0]*(xx - c[0]) + n[1]*(yy - c[1])) / (-n[2]) + c[2]
        >>> ax.plot_surface(xx, yy, zz, alpha=0.75, color="yellow")
        >>> plt.show(interactive=True)
        """
        assert points.shape[1] == 3
        return np.dot((points + 0.5) - self.centroid, self.normal_vector) >= 0



import re
from collections import Counter
from typing import Dict, List, Tuple

import numpy as np
import pandas as pd
from sklearn.decomposition import PCA


class Statistics:
    def __init__(self, measure_statistics_path):
        self.raw_statistics = pd.read_csv(measure_statistics_path)
        # self.all_namings, self.all_namings_no_duplicates, self.roi_naming_lookup = self._get_sorted_namings()

        self.nan_filtered_statistics = self._filter_empty_columns(self.raw_statistics)
        self.roi_statistics, self.dicom_metadata, self.spinal_canal_statistics = self._split_statistics(
            self.nan_filtered_statistics
        )
        self.statistics = self._get_statistics_without_rare_rois(self.roi_statistics)

        self.stats_per_roi = self.filter_by_column(
            self.statistics, excluded_substrings=["ANGLE_COR", "ANGLE_SAG", "HEIGHT_MIN"], reshape=True
        )

        self.centers = self.filter_by_column(self.statistics, ["CENTER"], reshape=True)

        # Calculate vectors

        self.disc_direction_vectors = self.filter_by_column(
            self.statistics, ["DIRECTION"], ["CANAL_DIRECTION"], reshape=True
        )
        self.canal_direction_vectors = self.filter_by_column(self.statistics, ["CANAL_DIRECTION"], reshape=True)
        # self.canal_direction_vectors = self.filter_by_column(
        #     self.statistics, ["VERTEBRA_CANAL_DIRECTION", "DISC_DIRECTION"], reshape=True
        # )
        self.roi_diff_vectors = self._calculate_roi_diff_vectors(self.centers)

        self.shape_printable = "\n".join(
            [
                f"\t{self.raw_statistics.shape=}",
                f"\t{self.statistics.shape=}",
                f"\t{self.centers.shape=}",
                f"\t{self.disc_direction_vectors.shape=}",
                f"\t{self.canal_direction_vectors.shape=}",
                f"\t{self.roi_diff_vectors.shape=}",
            ]
        )
        # s = Counter([c.split(";")[1] for c in full_statistics.columns])
        # print(s)

        # patient_centers = pca_normalise(patient_centers)

    # ===================================================
    # Public
    # ===================================================

    @property
    def all_vectors(self):
        return [self.disc_direction_vectors, self.canal_direction_vectors, self.roi_diff_vectors]

    @property
    def roi_naming_lookup(self):
        # vertebrae = [f"{c}{i+1}" for c, count in zip("CTLS", [7, 12, 5, 1] for i in range(count))][1:]
        # discs = [f"{c1}-{c2[1:] if c1[0] == c2[0] else c2}" for ]
        c = "C2 C2-3 C3 C3-4 C4 C4-5 C5 C5-6 C6 C6-7 C7 C7-T1 "
        t = "T1 T1-2 T2 T2-3 T3 T3-4 T4 T4-5 T5 T5-6 T6 T6-7 T7 T7-8 T8 T8-9 T9 T9-10 T10 T10-11 T11 T11-12 T12 T12-L1 "
        l = "L1 L1-2 L2 L2-3 L3 L3-4 L4 L4-5 L5 L5-S1 S1 "
        return dict(zip((c + t + l).split(), range(1000)))

    def filter_by_column(self, statistics=None, required_substrings=None, excluded_substrings=None, reshape=False):
        """Return shape (10k, 51 * #matching cols), if reshape==True, return shape (10k, 51, #matching cols)"""

        if statistics is None:
            statistics = self.statistics
        if required_substrings is None:
            required_substrings = statistics.columns
        if excluded_substrings is None:
            excluded_substrings = []

        relevant_columns = [
            column
            for column in statistics.columns
            if any(c in column for c in required_substrings) and all(c not in column for c in excluded_substrings)
        ]

        # namings = VectorAnnotator().roi_naming_lookup
        # sorted_columns = sorted(relevant_columns, key=lambda c: namings[c.split(";")[0]])

        combined_roi_xyz = statistics[relevant_columns]
        if reshape:
            separate_roi_xyz = self._reshape_combined_roi_xyz_to_separate_coordinates(combined_roi_xyz)
            return separate_roi_xyz
        return combined_roi_xyz

    def __str__(self):
        return "\nStatistics:\n" + self.shape_printable

    # ===================================================
    # Private
    # ===================================================

    def _split_statistics(self, statistics):
        """Split statistics into a list of statistics for each patient"""
        dicom_metadata = self.filter_by_column(statistics, ["DICOM;"])
        spinal_canal_statistics = self.filter_by_column(statistics, ["spinal_canal;"])

        used_columns = dicom_metadata.columns.tolist() + spinal_canal_statistics.columns.tolist()
        rest = self.filter_by_column(statistics, None, used_columns)
        return rest, dicom_metadata, spinal_canal_statistics

    def _filter_empty_columns(self, statistics):
        """Filter columns that are all NaN"""
        statistics = self.filter_by_column(statistics, None, ["unexpected_"])
        nan_filtered = statistics.dropna(axis=1, how="all")
        # nan_filtered = nan_filtered.dropna(axis=0, how="all")
        # nan_filtered = statistics.fillna(0.0)
        # assert nan_filtered.isna().sum().sum() == 0
        return nan_filtered

    def _get_statistics_without_rare_rois(self, measure_statistics):
        must_not_be_in = ["unexpected", "L1-S1", "L2-S1", "L3-S1", "spinal_canal"]
        statistics = self.filter_by_column(measure_statistics, None, must_not_be_in)

        centers = self.filter_by_column(statistics, ["CENTER"])

        skip_nan_percentage = 0.6
        column_nan_percentage = ((np.isnan(centers).sum(axis=0)) / centers.shape[0]) > skip_nan_percentage
        skip_rois = set(column.split(";")[0] for column in column_nan_percentage[column_nan_percentage].axes[0])
        remaining_columns = [column for column in statistics.columns if column.split(";")[0] not in skip_rois]

        sorted_columns = sorted(remaining_columns, key=lambda c: self.roi_naming_lookup.get(c.split(";")[0], 1e9))

        return statistics[sorted_columns]

    def _reshape_combined_roi_xyz_to_separate_coordinates(self, combined_roi_xyz):
        """Input shape: (10k, 153), output shape: (10k, 51, 3)"""

        remove_prefixes = "|".join(["DISC_", "VERTEBRA_"])
        # remove_prefixes = "|".join(["DISC_", "VERTEBRA_", "CANAL_"])

        counts = Counter([re.search(rf";({remove_prefixes})*(.*)", c).group(2) for c in combined_roi_xyz.columns])
        assert len(set(counts.values())) == 1, f"Column-sizes are not the same: {counts}"
        s = combined_roi_xyz.shape[1] // counts.most_common(1)[0][1]

        separate_roi_xyz = np.zeros((combined_roi_xyz.shape[0], combined_roi_xyz.shape[1] // s, s))
        for i in range(s):
            separate_roi_xyz[:, :, i] = combined_roi_xyz.iloc[:, i : combined_roi_xyz.shape[1] : s]
        return separate_roi_xyz

    def _pca_normalise(self, patient_centers):
        """Deprecated: use _sphere_coordinates_normalise instead"""

        pca = PCA(n_components=3)
        # pca = TruncatedSVD(n_components=3)
        # median_size = centers.mean(axis=1)
        for i in range(len(patient_centers)):
            if i % 1000 == 0:
                print(i)
            # current_centers = patient_centers[i]
            non_nan_centers = patient_centers[i][~np.isnan(patient_centers[i]).any(axis=1)]
            fitted = pca.fit(non_nan_centers)
            components = fitted.components_.T
            for j in range(3):
                if np.dot(components, np.array([1, 1, 1]))[j] < 0:
                    components[j] *= -1
            # patient_centers[i] -= fitted.mean_
            patient_centers[i] = patient_centers[i] @ components
            # current_centers -= fitted.mean_
            # print(fitted.components_)
            # current_centers /= fitted.components_[0]
        return patient_centers

    def _sphere_coordinates_normalise(self, patient_centers):
        # Shape: (10k, 51, 3)
        patient_centers -= patient_centers[:, -1][:, np.newaxis, :]

        # Flip axis because there were inconsistencies when some points were on the other side of the y axis
        x, y, z = 2, 0, 1

        # return patient_centers
        spherical = np.zeros(patient_centers.shape)
        spherical[:, :, x] = np.linalg.norm(patient_centers, axis=2)
        spherical[:, :, y] = np.arccos(patient_centers[:, :, z] / spherical[:, :, x])
        spherical[:, :, z] = np.arctan2(patient_centers[:, :, y], patient_centers[:, :, x])

        center_nth_roi = 0
        r = spherical[:, center_nth_roi, x]
        theta = spherical[:, center_nth_roi, y]
        phi = spherical[:, center_nth_roi, z]

        spherical[:, :, x] /= np.nan_to_num(r)[:, np.newaxis]
        # spherical[:, :, 0] = 10

        spherical[:, :, y] -= theta[:, np.newaxis] - np.pi / 2
        spherical[:, :, z] -= phi[:, np.newaxis]
        # spherical[:, :, 1] += (spherical[:, :, 2] > np.pi) * np.pi
        # spherical[:, :, 2] -= (spherical[:, :, 2] > np.pi) * np.pi

        # spherical[:, :, y] += .1

        new_centers = np.zeros(patient_centers.shape)
        new_centers[:, :, x] = spherical[:, :, x] * np.sin(spherical[:, :, y]) * np.cos(spherical[:, :, z])
        new_centers[:, :, y] = spherical[:, :, x] * np.sin(spherical[:, :, y]) * np.sin(spherical[:, :, z])
        new_centers[:, :, z] = spherical[:, :, x] * np.cos(spherical[:, :, y])

        return new_centers

    @staticmethod
    def predict_next_center_point(centers):
        """Input shape: (10k, 51, 3), output shape: (10k, 51+1, 3)"""

        x, y, z = centers[:, :, 0], centers[:, :, 1], centers[:, :, 2]

        # x_axis = np.arange(x.shape[1])[np.newaxis].repeat(x.shape[0], axis=0)
        x_axis = np.arange(x.shape[1])

        def fit(input_axis):
            optimal_params = np.polyfit(x_axis, input_axis.T, 2)
            poly = np.polyval(optimal_params, -1)
            return poly

        next_x, next_y, next_z = fit(x), fit(y), fit(z)
        new_centers = np.stack([next_x, next_y, next_z], axis=1)[:, None, :]
        return np.concatenate([centers, new_centers], axis=1)

    def _calculate_roi_diff_vectors(self, centers):
        centers[np.isnan(centers)] = np.mean(centers[~np.isnan(centers)])

        centers_with_new_top = self.predict_next_center_point(centers)
        vectors = centers_with_new_top[:, :-1] - centers_with_new_top[:, 1:]

        # normalise
        vectors = vectors / np.linalg.norm(vectors, axis=-1, keepdims=True)
        return vectors

    def _get_sorted_namings(self, discard_below_percentage=0.01) -> Tuple[List[str], List[str], Dict[str, int]]:
        # All namings sorted
        all_namings = set()
        counts = {}
        for column in self.raw_statistics.columns:
            if any(part in column for part in ["CENTER"]):
                roi_name = column.split(";")[0]
                all_namings.add(roi_name)
                counts[roi_name] = self.raw_statistics[column].count()
        all_namings -= {"spinal_canal", "unexpected_0"}
        all_namings = [n for n in all_namings if counts[n] / len(self.raw_statistics) > discard_below_percentage]
        all_namings += {"S1"}
        [(n, counts[n] / len(self.raw_statistics)) for n in all_namings]
        all_namings_sorted = []
        for current_part_of_spine in "CTLS":
            sort_by = lambda name: int(name[1:].split("-")[0]) + 0.5 * ("-" in name)
            all_namings_sorted.extend([n for n in sorted(all_namings, key=sort_by) if current_part_of_spine == n[0]])

        # Roi naming lookup
        roi_naming_lookup: Dict[str, int] = {}
        vertebra_index = 0
        for naming in all_namings_sorted:
            roi_naming_lookup[naming] = vertebra_index - ("-" in naming)
            vertebra_index += 2 * ("-" not in naming)

        # All namings no duplicates
        naming_occ = Counter(roi_naming_lookup.values())
        to_be_removed = [n for n in all_namings if naming_occ[roi_naming_lookup.get(n, 0)] >= 2 and "-S1" in n]
        all_namings_no_duplicates = list(filter(lambda n: n not in to_be_removed, all_namings_sorted))
        return all_namings_sorted, all_namings_no_duplicates, roi_naming_lookup



import numpy as np


class Spine:
    def __init__(self, data: np.ndarray, segmentation: np.ndarray):
        assert data.shape == segmentation.shape
        assert data.ndim == 3
        self.data = data
        self.segmentation = segmentation



from sklearn.ensemble import RandomForestClassifier

from spine_segmentation.datasets.feature_dataset import FeatureDataModule


def main():
    feature_data = FeatureDataModule(batch_size=8000, sample_size=30)

    train_loader = feature_data.train_dataloader()
    X, Y = next(iter(train_loader))
    X = X.view(X.shape[0], -1)
    Y -= 1

    val_loader = feature_data.val_dataloader()
    X_val, Y_val = next(iter(val_loader))
    X_val = X_val.view(X_val.shape[0], -1)
    Y_val -= 1
    print("Train size", len(X), "Val size", len(X_val))

    # X, Y = [], []
    # for x, y in train_loader:
    #     X.append(x)
    #     Y.append(y)

    # Train a random forrest classifier
    # from sklearn.metrics import plot_confusion_matrix
    # from sklearn.metrics import plot_roc_curve
    # from sklearn.metrics import plot_precision_recall_curve
    # from sklearn.metrics import roc_auc_score
    import xgboost as xgb
    from sklearn.metrics import accuracy_score, classification_report, confusion_matrix

    # Use XGBoost
    # clf = sklearn.ensemble.GradientBoostingClassifier(n_estimators=1000, random_state=0)
    clf = xgb.XGBClassifier()

    # Use random forest
    clf = RandomForestClassifier(n_estimators=1000, random_state=0)

    clf.fit(X, Y)

    y_pred = clf.predict(X)
    print("Train Accuracy:", accuracy_score(Y, y_pred))

    # Evaluate the model
    y_pred = clf.predict(X_val)
    print("Accuracy:", accuracy_score(Y_val, y_pred))
    print("Confusion matrix:")
    print(confusion_matrix(Y_val, y_pred))
    print("Classification report:")
    print(classification_report(Y_val, y_pred))


if __name__ == "__main__":
    main()



import logging
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, List, Tuple

import matplotlib
import numpy as np
from matplotlib import pyplot as plt

from spine_segmentation.annotators.annotator_base import AnnotatorSimple
from spine_segmentation.spine_types.disc import Disc
from spine_segmentation.spine_types.roi import Roi
from spine_segmentation.spine_types.spine import Spine

logger = logging.getLogger(__name__)


@dataclass
class _SubsetErrorScore:
    score: float
    index: int
    subset: np.ndarray
    naming: List[str]

    def __lt__(self, other):
        return self.score < other.score


# _SubsetErrorScore = namedtuple("_SubsetErrorScore", ["score", "index", "subset", "naming"])
def _plot_error_for_each_start_index(error_scores: List[_SubsetErrorScore], size: int, correct: int, debug_dir: Path):
    best, *others = sorted(error_scores)
    second_best = _SubsetErrorScore(best.score / 10, -1, None, None) if len(others) == 0 else others[0]
    index_to_score = defaultdict(lambda: 1e100)
    for error_score in error_scores:
        index_to_score[error_score.index] = min(index_to_score[error_score.index], error_score.score)
    starts = sorted(index_to_score)
    scores = list(zip(*sorted(index_to_score.items())))[1]
    certainty = 1 - best.score / second_best.score
    color_lookup = defaultdict(lambda: ("b", "Any"))
    color_lookup.update(
        {
            best.index: ("r", "Best guess (incorrect)"),
            second_best.index: ("y", "Second best guess"),
            correct: ("g", "Actual"),
        }
    )
    colors = [color_lookup[k][0] for k in starts]
    plt.bar(starts, scores, color=colors)
    plt.xticks(starts)
    color_and_label = set(color_lookup.values())
    plt.legend(handles=[matplotlib.patches.Patch(color=col, label=label) for col, label in color_and_label])
    plt.xlabel("ROI start index")
    plt.ylabel("Error (smallest is best)")
    plt.title(f"Error for sample of {size=}, {certainty=:.1%} {'CORRECT' if correct == best.index else 'WRONG'}")
    plt.savefig(debug_dir / "annotator.png")
    plt.cla()


class VectorAnnotator(AnnotatorSimple):
    def _find_best_match(self, table, sample, is_disc: bool = None, correct=-1, debug_dir: Path = None) -> List[str]:
        size = len(sample)
        error_scores: List[_SubsetErrorScore] = []
        earliest_possible_start = 0 if is_disc is None else int(is_disc)
        latest_possible_start = table.shape[1] - size + 1
        scoring_function = self._get_scoring_function()
        for start in range(earliest_possible_start, latest_possible_start, 1 if is_disc is None else 2):
            for subset, naming in self._get_subsets_at(table, start, size):
                error_scores.append(_SubsetErrorScore(scoring_function(subset, sample), start, subset, naming))
        if debug_dir is not None:
            _plot_error_for_each_start_index(error_scores, size, correct, self.debug_dir)
        best_error_score = sorted(error_scores)[0]
        return best_error_score.naming

    def _plot_vectors_for_best_guess(self, table, error_scores) -> List:
        pass

    @staticmethod
    def _mse_sample_on_subset(sample, subset):
        if len(sample.shape) != 3:
            raise Exception(f"{len(sample.shape)=} != 3")
        if len(subset.shape) != 4:
            raise Exception(f"{len(subset.shape)=} != 4")
        diff = subset - sample  # e.g. shape: (8511, 21, 2, 3)
        return np.average(diff**2, axis=tuple(range(1, len(diff.shape))))  # e.g. shape: (8511)

    @staticmethod
    def _get_scoring_function(strategy="mse_top_n", n=50) -> Callable[[np.ndarray, np.ndarray], float]:
        """Returns a function which takes a subset and a sample and returns some error value

        Subset is of shape [#nako_patients, #rois, #properties, 3], e.g. [8511, 21, 2, 3]
        Sample is of shape [#rois, #properties, 3], e.g. [21, 2, 3]

        """

        mse_scaling = (np.cos(np.pi * np.arange(n) / (n - 1)) + 1) / 2.5 + 0.2

        def mse_top_n(subset, sample) -> float:
            mse = VectorAnnotator._mse_sample_on_subset(sample, subset)  # e.g. shape: (8511)
            min_values = np.sort(np.partition(mse, n)[:n]) * mse_scaling  # e.g. shape: (50)
            return np.average(min_values)  # e.g. shape: (1)

        def naive_all_sum(subset, sample) -> float:
            diff = subset - sample
            return np.log(np.sum(diff**2))

        return {
            "mse_top_n": mse_top_n,
            "naive_all_sum": naive_all_sum,
        }[strategy]

    def _ensure_alternating_rois(self, rois: List[Roi]):
        """Ensures that the rois are alternating between disc and vertebra

        For this it creates a new roi if there are two consecutive rois of the same type.
        The other rois are unchanged, and their index would be incorrect.
        Sometimes there is a missing roi, which actually exists but was to segmented correctly.
        The missing roi is interpolated from the two surrounding rois of the same type (if this is not possible then
        any adjacent 2 rois are used)
        """

        @dataclass
        class FakeRoi:
            direction: List[float] = None
            volume: float = None

        i = 1
        while i < len(rois):
            # Check if the rois are of the same class
            above, below = rois[i - 1], rois[i]
            above2, below2 = above, below
            if 0 <= i - 2 <= i + 1 < len(rois):
                above2, below2 = rois[i - 2], rois[i + 1]
            if above.__class__ == below.__class__:
                # If so, create a new FakeRoi and insert it there
                direction = (above.direction + below.direction) / 2
                direction = direction / np.linalg.norm(direction)
                volume = (above2.volume + below2.volume) / 2
                rois.insert(i, FakeRoi(direction, volume))
            i += 1

    def _make_vector_from_spine(self, spine) -> Tuple[np.ndarray, bool]:
        """Returns shape (#rois, #properties, 3) and whether the first roi is a disc"""
        rois = spine.get_rois(sort=True, skip_canal=True).copy()
        self._ensure_alternating_rois(rois)
        rois_as_vectors = np.zeros((len(rois), len(self.properties_to_be_used), 3))
        for roi_index, roi in enumerate(rois):
            for index, measure_name in enumerate(self.properties_to_be_used):
                rois_as_vectors[roi_index, index, :] = getattr(roi, measure_name.lower())
        is_disc = isinstance(rois[0], Disc)
        return rois_as_vectors, is_disc

    # def _apply_new_spine_roi_naming(self, spine: Spine, naming: List[str]):
    #     rois = spine.get_rois(sort=True, skip_canal=True)
    #     if len(rois) != len(naming):
    #         logger.warning(f"There is likely a roi missing in the spine {spine.uid}, interpolating it for naming!")
    #     roi_index = 0
    #     for naming_index in range(len(naming)):
    #         roi = rois[roi_index]
    #         if isinstance(roi, Disc) and "-" in naming[naming_index]:
    #             roi.name = naming[naming_index]
    #             roi_index += 1
    #         elif isinstance(roi, Vertebra) and "-" not in naming[naming_index]:
    #             roi.name = naming[naming_index]
    #             roi_index += 1
    #     # for roi, name in zip(rois, naming):
    #     #    roi.name = name
    #     logger.info(f"sort rois: {[r.name for r in rois]}")

    def _process(self, spine: Spine):
        if self._is_naming_incomplete(spine):
            logger.info(f"Naming is incomplete for patient {spine.uid}")
            spine_vector, is_disc = self._make_vector_from_spine(spine)
            spine_vector = self._normalize(spine_vector)
            naming_proposal = self._find_best_match(self.vector_table, spine_vector, is_disc=is_disc)
            if naming_proposal[-1] in ("L6", "L5", "L4"):
                naming_proposal[-1] = "S1"
                naming_proposal[-2] = f"{naming_proposal[-3]}-S1"
            self._apply_new_spine_roi_naming(spine, naming_proposal)


if __name__ == "__main__":
    annotator = VectorAnnotator()
    annotator._algorithm_validation(iterations=1000)



from typing import Literal

import torch
from kornia import augmentation as K
from torch import Tensor


class ImageAugmentation(torch.nn.Module):
    """Module to perform data augmentation using Kornia on torch tensors."""

    def __init__(
        self,
        cropping_type: Literal["none", "same", "many"] = "many",
        geometric_aug_type: Literal["none", "all"] = "all",
        photometric_aug_type: Literal["none", "all"] = "all",
        use_mask=True,
    ) -> None:
        super().__init__()

        self.use_mask = use_mask

        # === GEOMETRIC ===
        geometric_augmentations = []

        if geometric_aug_type == "all":
            geometric_augmentations += [
                K.RandomVerticalFlip(),  # Note that patient is lying on back, so this is actually a horizontal flip
                K.RandomAffine(degrees=20, translate=(0.1, 0.1), scale=(0.8, 1.2), p=0.50),
            ]

        if cropping_type == "same":
            geometric_augmentations += [
                K.RandomCrop((320, 300), p=1.0),
                K.PadTo((320, 896)),
            ]
        if cropping_type == "many":
            geometric_augmentations += [
                K.RandomCrop((320, 300), p=0.1),
                K.RandomCrop((320, 400), p=0.15),
                K.RandomCrop((320, 500), p=0.2),
                K.RandomCrop((320, 600), p=0.2),
                K.RandomCrop((320, 700), p=0.2),
                K.PadTo((320, 896)),
            ]

        self.geometric_augmentations = K.AugmentationSequential(
            *geometric_augmentations, data_keys=(["input", "mask"] if use_mask else ["input"])
        )

        # === PHOTOMETRIC ===
        photometric_augmentations = []
        if photometric_aug_type == "all":
            photometric_augmentations += [
                K.RandomBrightness((0.9, 1.1), p=0.25),
                K.RandomContrast((0.9, 1.1), p=0.25),
                K.RandomGaussianBlur((3, 3), (0.1, 2.0), p=0.25),
                K.RandomGaussianNoise(0.0, 0.01, p=0.25),
                K.RandomGaussianNoise(0.0, 0.02, p=0.15),
                K.RandomGaussianNoise(0.0, 0.03, p=0.05),
            ]

        self.photometric_augmentations = K.AugmentationSequential(
            # K.RandomGamma((0.9, 1.1), (1.5, 1.5), p=0.5),
            data_keys=["input"],
        )

    @torch.no_grad()
    def forward(self, inputs: Tensor, mask: Tensor = None):
        if self.use_mask:
            inputs, mask = self.geometric_augmentations(inputs, mask)
        else:
            inputs = self.geometric_augmentations(inputs)
        inputs = self.photometric_augmentations(inputs)
        if self.use_mask:
            return inputs, mask
        return inputs

    def single_image(self, inputs: Tensor, mask: Tensor):
        inputs, mask = self.forward(inputs[None], mask[None])
        return inputs[0], mask[0]



import subprocess
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, List, Literal, Tuple, Union

import numpy as np

from spine_segmentation.spine_types.line_segment import LineSegment
from spine_segmentation.spine_types.plane import Plane
from spine_segmentation.visualisation.blender.gen_obj import generate_obj, normalize
from spine_segmentation.visualisation.color import Color
from spine_segmentation.visualisation.color_palettes import get_alternating_palette

_geom_scale = (3.3, 0.8, 0.8)
_geom_transform_mat = np.array(
    [
        [0, _geom_scale[1], 0, 0],
        [0, 0, _geom_scale[2], 0],
        [_geom_scale[0], 0, 0, 0],
        [0, 0, 0, 1],
    ]
)


class BlenderCollection:
    def __init__(self, name: str, objects: List = [], *, hidden: bool = False):
        self.name = name
        self.hidden = hidden
        self.objects = objects

    def script_args(self):
        hidden_str = "-hidden" if self.hidden else ""
        return [f"--collection{hidden_str}", self.name]


def create_3d_models_and_labels(
    spine, *, directory: Union[str, Path], color_mask
) -> Tuple[List[Path], Dict[str, Tuple]]:
    """Creates the 3D .obj files for the spine itself, filters rois, such that only ones in the mask are included"""
    directory = Path(directory)
    directory.mkdir(exist_ok=True)
    paths = []

    # def mask_overlaps_with(roi):
    #     """Return True if there is any overlap between the mask and the given roi"""
    #     roi_bounds = [self._index3D_to_image2D(p) for p in roi.boundingbox_raw]
    #     roi_bounds_min, roi_bounds_max = np.min(roi_bounds, axis=0), np.max(roi_bounds, axis=0)
    #     mask_bounds_min, mask_bounds_max = self._mask_cropping_box[:2], self._mask_cropping_box[2:]
    #     bounds_min = np.max([mask_bounds_min, roi_bounds_min], axis=0)
    #     bounds_max = np.min([mask_bounds_max, roi_bounds_max], axis=0)
    #     return all((bounds_max - bounds_min) > 0)

    # filtered_vertebrae = list(filter(mask_overlaps_with, self._spine.vertebras))
    # filtered_discs = list(filter(mask_overlaps_with, self._spine.discs))

    name_and_id_set_pairs = [
        ("Vertebrae", spine, set(range(1, 100, 2))),
        ("Discs", spine, set(range(2, 100, 2))),
        # ("Spinal_canal", {self.spine.idx_spinal_canal}, (self._spine.seg_canal != 0) * self.spine.idx_spinal_canal),
    ]
    for name, seg_3d, id_set in name_and_id_set_pairs:
        obj_path = directory / f"{name}.obj"
        print(f"Generating {obj_path} {seg_3d.shape}")
        # seg_3d = seg_3d[::2, ::2, ::2]
        generate_obj(
            obj_path,
            seg_3d,
            id_set,
            seg_3d,
            type_to_color=color_mask,
            rot_mat=_geom_transform_mat,
        )
        paths.append(obj_path)

    return paths


def _write_obj_if_possible_for_object(obj: object, filepath: Path) -> bool:
    """Create a .obj file for the given object if possible, return True if yes"""

    def write_obj_file(points, face_or_line: Literal["f", "l"], start_at=1, rot_mat=_geom_transform_mat):
        points = normalize(points, rot_mat=rot_mat)
        with open(filepath, "a") as obj_file:
            for point in points:
                obj_file.write("v " + " ".join(map(str, point)) + "\n")
            obj_file.write(f"{face_or_line} " + " ".join(map(str, range(start_at, len(points) + start_at))) + "\n")

    if isinstance(obj, LineSegment):  # or isinstance(obj, Polygon):
        write_obj_file(obj.points, "l")
    if isinstance(obj, Plane):
        write_obj_file(obj.points, "f")
    if isinstance(obj, np.ndarray):
        if len(obj.shape) == 3:
            indices_for_3D = set(np.unique(obj)) - {0}  # Do not use empty points for constructing the object
            generate_obj(filepath, obj, indices_for_3D, obj, rot_mat=_geom_transform_mat)
    # if isinstance(obj, Ellipsoid):
    #     generate_ellipsoid_obj(filepath, obj, self._spine.seg.shape, self._geom_transform_mat)
    # if isinstance(obj, AxisParallelCuboid):
    #     for index, face in enumerate(obj.faces):
    #         write_obj_file(np.array(face), "f", start_at=1 + index * 4)
    return filepath.exists()


def open_in_blender(
    spines: Union[np.ndarray, Dict[str, np.ndarray]],
    collections: List[BlenderCollection] = [],
    *,
    smooth_spine: bool = True,
    export_to_instead: Union[bool, str, Path] = False,
    render_to_instead: Union[bool, str, Path] = False,
):
    """Opens a 3D view in blender of all added objects. Useful for debugging. Blender 2.80+ is required.
    Pauses program until the blender process finishes.

    @param spines: the spine to show, 3D numpy array
    @param collections: list of blender collections which will be rendered additionally
    @param smooth_spine: whether a smoothing modifier should be applied to spine, otherwise raw voxels are shown
    @param export_to_instead: instead of opening Blender, export a .dae model to the given path. If True simply
           export to the temp directory with the name "model.dae".
    @param render_to_instead: instead of opening Blender, render the scene to the given path. If True simply
           render to the temp directory with the name "render.png".
    @return: None
    """

    alternating_palette = get_alternating_palette()

    prefix_to_color_matching = {
        "instances": alternating_palette,
        "output": {1: Color("orange"), 2: Color("lightgreen")},
        "spine": {1: Color("orange"), 2: Color("lightgreen")},
        "gt": {1: Color("red"), 2: Color("black")},
    }

    if not isinstance(spines, dict):
        spines = {"spine": spines}
    with TemporaryDirectory(prefix="spine_blender_") as tmp_dir:
        script_path = Path(__file__).absolute().parent / "blender_script.py"
        blender_args = ["blender", "-P", str(script_path)]
        if export_to_instead or render_to_instead:
            blender_args.extend(["--background"])

        script_args = []
        for prefix, spine in spines.items():
            script_args.extend(["--collection", prefix.capitalize()])
            color_mask = prefix_to_color_matching.get(prefix, None)
            obj_paths = create_3d_models_and_labels(spine, directory=Path(tmp_dir) / prefix, color_mask=color_mask)
            for obj_path in obj_paths:
                spine_arg = "--object-smooth" if smooth_spine else "--object"
                script_args.extend([spine_arg, str(obj_path)])

        if render_to_instead:
            path = Path(tmp_dir) / "render.png" if render_to_instead is True else Path(render_to_instead)
            path.parent.mkdir(parents=True, exist_ok=True)
            script_args.extend(["--render-to", str(path)])

        for collection in collections:
            script_args.extend(collection.script_args())
            collection_dir = Path(tmp_dir) / collection.name
            collection_dir.mkdir(parents=True)

            for index, object_3D in enumerate(collection.objects, 1):
                filename = getattr(object_3D, "name", f"{collection.name[:-1]}{index}")
                object_3D_path = collection_dir / f"{filename}.obj"
                print(collection, object_3D_path)
                if _write_obj_if_possible_for_object(object_3D, object_3D_path):
                    script_args.extend(["--object", str(object_3D_path)])

        if export_to_instead:
            path = Path(tmp_dir) / "model.dae" if render_to_instead is True else Path(render_to_instead)
            path.parent.mkdir(parents=True, exist_ok=True)
            # Export the blender composed model
            script_args.extend(["--export-to", str(path)])
        #     # Create the json file with the center names
        #     vert_centers_path = export_to_instead.with_name("vertebrae_centers.json")
        #     if not vert_centers_path.exists():
        #         with open(vert_centers_path, "w") as file:
        #             import json
        #             json.dump(center_labels, file)

        arg_separator = ["--"]
        print(blender_args + arg_separator + script_args)
        subprocess.run(blender_args + arg_separator + script_args)



import time
from contextlib import contextmanager

profiling_enabled = True


def profile(fn):
    def wrapper(*args, **kwargs):
        if profiling_enabled:
            start_time = time.time()
        result = fn(*args, **kwargs)
        if profiling_enabled:
            elapsed_time = time.time() - start_time
            print(f"'{fn.__name__}': elapsed time={elapsed_time:.4f} seconds")
        return result

    return wrapper


@contextmanager
def profile_context(name):
    if profiling_enabled:
        start_time = time.time()
    yield
    if profiling_enabled:
        elapsed_time = time.time() - start_time
        print(f"{name}: elapsed time={elapsed_time:.4f} seconds")



import datetime
from functools import lru_cache
from pathlib import Path


@lru_cache(maxsize=1)
def get_next_log_dir():
    log_dir = Path(__file__).parent.parent.parent / "logs"
    log_dir = log_dir / datetime.datetime.now().strftime("%Y-%m-%d_%H%M%S")
    log_dir.mkdir(exist_ok=True, parents=True)
    return log_dir


@lru_cache(maxsize=1)
def get_next_log_dir_on_tank():
    log_dir = Path("<log_dir>")
    log_dir = log_dir / datetime.datetime.now().strftime("%Y-%m-%d_%H%M%S")
    log_dir.mkdir(exist_ok=True, parents=True)
    return log_dir



from dataclasses import dataclass


@dataclass
class Number:
    number: int = 0

    def get(self):
        return self.number

    def set(self, number):
        self.number = number


TRAIN_DATASET_SIZE = Number(0)



import torch


class DiffIs1Loss(torch.nn.Module):
    def forward(self, prediction, _):
        prediction = prediction.squeeze()
        mult_factor = 3
        diff = torch.abs((prediction[1:] - 1) - prediction[:-1])
        pow = torch.pow(diff, 2)
        loss = torch.mean(pow)
        return loss



from typing import Tuple, TypeVar, Union

import numpy as np


class Color:
    colors = {
        "grey": (128, 128, 128, 255),
        "red": (255, 0, 0, 255),
        "lightred": (255, 144, 144, 255),
        "green": (0, 255, 0, 255),
        "lightgreen": (144, 238, 144, 255),
        "darkgreen": (0, 100, 0, 255),
        "blue": (0, 0, 255, 255),
        "yellow": (255, 255, 0, 255),
        "orange": (255, 165, 0, 255),
        "brown": (165, 42, 42, 255),
        "cyan": (0, 255, 255, 255),
        "magenta": (255, 0, 255, 255),
        "violet": (238, 130, 238, 255),
        "purple": (128, 0, 128, 255),
        "black": (0, 0, 0, 255),
        "white": (255, 255, 255, 255),
        "transparent": (255, 255, 255, 0),
    }
    ANSI_RESET = "\033[0m"
    ANSI_BOLD = "\033[1m"

    def __init__(
        self,
        red_or_grey_or_color: "ColorLike",
        green: Union[float, int, None] = None,
        blue: Union[float, int, None] = None,
        /,
        alpha: Union[float, int, None] = None,
    ):
        """Color either as 1 (grey), 3 (RGB), 4 (RGBA) numbers, a string ("red") or hexcode ("#ff0000")

        Can be in range of either [0.0, 1.0] if floats or [0, 255] if ints. Must be all same type.
        If alpha not supplied, it is implied to be 255 (fully opaque).
        If first argument is str, the other arguments are ignored.
        """
        if isinstance(red_or_grey_or_color, Color):
            red, green, blue, alpha = red_or_grey_or_color.ints()
        elif isinstance(red_or_grey_or_color, tuple):
            red, green, blue, alpha = Color(*red_or_grey_or_color).ints()
        elif isinstance(red_or_grey_or_color, str):
            red, green, blue, alpha = self._from_str(red_or_grey_or_color)
        else:
            red_or_grey = red_or_grey_or_color
            if green is None and blue is None:
                green = blue = red_or_grey
            elif green is None or blue is None:
                raise TypeError(f"One of green={green} or blue={blue} is supplied, but not both!")
            red = red_or_grey
            if alpha is None:
                alpha = 1.0 if isinstance(red, float) else 255
        if not (type(red) is type(green) is type(blue) is type(alpha)):
            raise TypeError(
                f"Color variables not of the same type: "
                f"{type(red)} != {type(green)} != {type(blue)} != {type(alpha)}"
            )
        self._color: Tuple[int, int, int, int] = red, green, blue, alpha
        if isinstance(red, float):
            self._color = tuple(map(lambda c: int(c * 255), self._color))
        assert len(self._color) == 4
        if not all(0 <= c <= 255 for c in self._color):
            raise ValueError(f"Color out of bounds: {self._color} (must be between 0 and 255)")

    def ints(self) -> Tuple[int, int, int, int]:
        return self._color

    def floats(self) -> Tuple[float, float, float, float]:
        return tuple(c / 255 for c in self._color)

    def hex(self) -> str:
        return "#" + "".join(f"{c:02x}" for c in self._color)

    def inverted(self) -> "Color":
        return Color(255 - self.r, 255 - self.g, 255 - self.b, self.a)

    def ansi(self) -> str:
        return f"\033[38;2;{';'.join(map(str, self._color))}m"

    def ansi_surround(self, text: str, bold: bool = False) -> str:
        additional = ""
        if bold:
            additional += self.ANSI_BOLD
        return additional + self.ansi() + text + self.ANSI_RESET

    def hsv(self) -> Tuple[float, float, float, float]:
        # Based on conversion from https://en.wikipedia.org/wiki/HSL_and_HSV
        r, g, b, alpha = self.floats()
        color_max, color_min = max(r, g, b), min(r, g, b)
        delta = color_max - color_min
        if delta == 0:
            hue = 0
        elif color_max == r:
            hue = ((g - b) / delta) % 6
        elif color_max == g:
            hue = ((b - r) / delta) + 2
        elif color_max == b:
            hue = ((r - g) / delta) + 4
        hue = 60.0 * hue
        # angle_alpha = 0.5 * (2 * r - g - b)
        # angle_beta = np.sqrt(3) * 2 * (g - b)
        # hue_angle = np.atan2(angle_beta, angle_alpha)
        value = color_max
        saturation = 0 if value == 0 else delta / value
        return hue, saturation, value, alpha

    @staticmethod
    def from_hsv(hue: float, saturation: float, value: float, alpha: float = 1.0) -> "Color":
        """Create a Color object from a hsv value

        @param hue: angle on the color wheel hue [0.0, 360.0)
        @param saturation: percentage of saturation [0.0, 1.0]
        @param value: percentage for brightness [0.0, 1.0]
        @param alpha: percentage opaqueness [0.0, 1.0] (0 is fully transparent, 1 is fully opaque)
        @return: Color object
        """
        hue %= 360
        c = value * saturation
        x = c * (1 - abs((hue / 60) % 2 - 1))
        m = value - c
        rgb = [(c, x, 0), (x, c, 0), (0, c, x), (0, x, c), (x, 0, c), (c, 0, x)][int(hue // 60)]
        r, g, b = map(float, tuple((np.array(rgb) + m)))
        return Color(r, g, b, alpha)

    @staticmethod
    def _calculate_alpha_interpolation_factors(arr, interpolate_from, take_new_factor):
        alpha = 3
        with np.errstate(divide="ignore", invalid="ignore"):
            new_factors = (
                take_new_factor * interpolate_from[..., alpha] / (arr[..., alpha] + interpolate_from[..., alpha])
            )
        np.nan_to_num(new_factors, copy=False, nan=1)
        new_factors = new_factors[..., np.newaxis]
        return new_factors, 1 - new_factors

    @staticmethod
    def make_gradient(color1, color2, number_of_values):
        # which takes 2 numbers as tuples of 4 floats and creates a list of number of values which are slowly interpolated from color1 to color2
        # color1 and color2 are in rgb format
        take2_factor = np.arange(number_of_values)[:, None] / number_of_values
        take1_factor = 1 - take2_factor
        gradient = color1.floats() * take1_factor + color2.floats() * take2_factor
        return [Color(*c) for c in gradient]

    @staticmethod
    def fast_interpolate_rgb_arrays(arr, interpolate_from, take_new_factor=0.6):
        """
        Expected shape for arrays: [x, y, 4], where the last dimension is red, green, blue, alpha.
        Hue is in degrees from 0 to 360, saturation, value and alpha are each between 0 and 1
        @param arr:
        @param interpolate_from:
        @param take_new_factor:
        @return:
        """
        assert arr.shape[-1] == 4 == interpolate_from.shape[-1], "Arrays do not contain color data"
        new_factors, old_factors = Color._calculate_alpha_interpolation_factors(arr, interpolate_from, take_new_factor)
        arr = arr * old_factors + interpolate_from * new_factors
        return arr

    @staticmethod
    def fast_interpolate_hsv_arrays(arr, interpolate_from, take_new_factor=0.6):
        """
        Expected shape for arrays: [x, y, 4], where the last dimension is red, green, blue, alpha.
        Hue is in degrees from 0 to 360, saturation, value and alpha are each between 0 and 1
        @param arr:
        @param interpolate_from:
        @param take_new_factor:
        @return:
        """
        assert arr.shape[-1] == 4 == interpolate_from.shape[-1], "Arrays do not contain color data"
        hue, saturation, value, alpha = 0, 1, 2, 3
        new_factors, old_factors = Color._calculate_alpha_interpolation_factors(arr, interpolate_from, take_new_factor)
        # Find the shortest angle between the 2 angles
        # proof: https://math.stackexchange.com/questions/2144234/interpolating-between-2-angles
        # The 540 was replaced with 180, because in python modulo of negative numbers returns positive numbers,
        # e.g.: -400 % 360 = 320 (instead of -40, like in many other programming languages). So a 540 is unnecessary
        # here, as it as used to guarantee that the number is positive
        angles_between = ((interpolate_from[..., hue] - arr[..., hue]) % 360 + 180) % 360 - 180
        arr[..., hue] = (arr[..., hue] + angles_between * new_factors[..., 0]) % 360
        sat_alpha = slice(saturation, alpha + 1)
        arr[..., sat_alpha] = arr[..., sat_alpha] * old_factors + interpolate_from[..., sat_alpha] * new_factors
        return arr

    @staticmethod
    def hsv_array_to_rgb_array(hsv_array: np.ndarray) -> np.ndarray:
        assert hsv_array.shape[-1] == 4, "Array does not contain color data in last axis (or does not contain alpha)"
        hue, value, saturation, alpha = (hsv_array[..., i] for i in range(4))
        hue %= 360
        chroma = value * saturation
        x = chroma * (1 - abs((hue / 60) % 2 - 1))
        m = value - chroma
        hue_as_index = (hue // 60).astype(np.uint8)
        hue_as_index = hue_as_index[np.newaxis, ...]
        zeros = np.zeros_like(hue)
        red = np.take_along_axis(np.stack([chroma, x, zeros, zeros, x, chroma]), hue_as_index, axis=0)[0]
        green = np.take_along_axis(np.stack([x, chroma, chroma, x, zeros, zeros]), hue_as_index, axis=0)[0]
        blue = np.take_along_axis(np.stack([zeros, zeros, x, chroma, chroma, x]), hue_as_index, axis=0)[0]
        rgba_stacked = np.stack([red, green, blue, alpha], axis=-1)
        mmm_no_alpha = np.stack([m, m, m, np.zeros_like(m)], axis=-1)
        # Add m value to rgb, but not to alpha
        rgba_array = ((rgba_stacked + mmm_no_alpha) * 255).round().astype(np.uint8)
        return rgba_array

    @staticmethod
    def _from_hex(color_hex: str) -> Tuple[int, int, int, int]:
        """Create color from hex code such as #af22e3 (RGB) or #3545bf5f (RGBA), # is optional"""
        color_hex = color_hex.lstrip("#")
        if len(color_hex) == 3:
            color_hex = color_hex[0] * 2 + color_hex[1] * 2 + color_hex[2] * 2
        if len(color_hex) == 6:
            color_hex += "FF"
        import re

        if not re.fullmatch(r"#?[0-9a-fA-F]{8}", color_hex):
            raise ValueError(f"Invalid or unknown color hex code: {color_hex}!")
        assert len(color_hex) == 8
        colors = tuple(int(color_hex[i : i + 2], 16) for i in range(0, len(color_hex), 2))
        return colors

    @staticmethod
    def _from_str(color_str: str) -> Tuple[int, int, int, int]:
        # Assume transparency, similar to https://en.wikibooks.org/wiki/LaTeX/Colors
        # <COLOR>!transparency-percent (<COLOR>!100==<COLOR>, <COLOR>!0=="transparent"
        if "!" in color_str:
            color, percent = color_str.split("!", maxsplit=1)
            color_int = Color._from_str(color)
            try:
                return *color_int[:3], int(color_int[3] * int(percent) / 100)
            except ValueError:
                pass  # Will raise error at end of function
        try:
            return Color._from_hex(color_str)
        except ValueError:
            pass  # Will raise error at end of function
        if color_str.lower() in Color.colors:
            return Color.colors[color_str.lower()]
        raise ValueError(f"Unknown color_string: {color_str}")

    def __eq__(self, other: "Color"):
        return self._color == other.ints()

    def __str__(self):
        return "Color=" + str(self._color)

    def __repr__(self):
        return "Color=" + str(dict(zip("rgba", self._color)))

    @property
    def r(self) -> int:
        return self._color[0]

    @property
    def g(self) -> int:
        return self._color[1]

    @property
    def b(self) -> int:
        return self._color[2]

    @property
    def a(self) -> int:
        return self._color[3]

    @property
    def visible(self):
        return self.a > 0

    @classmethod
    def random(cls):
        from random import random

        return Color(random(), random(), random())


ColorLike = TypeVar(
    "ColorLike",
    int,
    Tuple[int, int, int],
    Tuple[int, int, int, int],
    float,
    Tuple[float, float, float],
    Tuple[float, float, float, float],
    Color,
    str,  # hex-code or string
)



from pathlib import Path
from typing import Union, List

import numpy as np

from spine_segmentation.resources.other_paths import RAW_NAKO_DATASET_PATH
from spine_segmentation.resources.paths import NAKO_DATASET_PATH


def expand_path_to_data_dirs(path: Union[Path, str]) -> List[Path]:
    path = Path(path)
    if path.suffix == ".npz":
        npz = np.load(path, allow_pickle=True)
        return list(npz["index_to_zip_path"].item().values())

    if path.suffix == ".csv":
        import pandas as pd

        csv = pd.read_csv(path)
        return list(csv["Path"])

    if path == RAW_NAKO_DATASET_PATH:
        return list(RAW_NAKO_DATASET_PATH.glob("*.zip"))

    if path == NAKO_DATASET_PATH:
        return get_potential_data_dirs(path, required_suffixes=("", "_WK"))

    raise ValueError(f"Invalid path {path=}")


def get_potential_data_dirs(data_dir, required_suffixes=("", "_BS", "_WK", "_NR"), intersection_with=None):
    if intersection_with is None:

        class AlwaysTrue:
            def __contains__(self, _):
                return True

        intersection_with = AlwaysTrue()
    paths = []
    for path in Path(data_dir).glob("*"):
        if path.name in intersection_with:
            if all((path / f"{path.name}{suffix}.nii.gz").exists() for suffix in required_suffixes):
                paths.append(path)
    return paths



import random

from spine_segmentation.cli.cli import colored_tracebacks
from spine_segmentation.datasets.sample import get_random_crop_slice
from spine_segmentation.evaluate.eval_instance_seg_model import post_process_instances
from spine_segmentation.inference.onnx_model import ONNXInferenceModel
from spine_segmentation.plotting.plot_slice import get_stats, plot_npz
from spine_segmentation.resources.paths import TRAIN_SPLIT_CSV_PATH, VAL_SPLIT_CSV_PATH
from spine_segmentation.utils.log_dir import get_next_log_dir

colored_tracebacks()

import numpy as np

from spine_segmentation.datasets.segmentation_dataset import SegmentationDataModule


def eval_every_seg_sample_by_splitting(directory=None, device=3, height=416):
    if directory is None:
        directory = get_next_log_dir()

    onnx_seg_inference = ONNXInferenceModel.get_best_segmentation_model(device=device)
    # onnx_seg_inference.load_index_list()

    # onnx_inst_inference = ONNXInferenceModel.get_best_instance_segmentation_model(px=height, device=device)

    # sample_iterator = SampleIterator(
    #    expand_path_to_data_dirs(RAW_NAKO_DATASET_PATH), add_adjacent_slices=True, skip_first_percent=0.9505
    # )

    # index_list_path = (
    #    Path.home() / "devel/src/git/spine_annotator/cache/onnx_inference/2023-09-12_Seg_Unet_mitb5/index_list.npz"
    # )
    # val_dataloader = get_dataloader("val", index_list_path, bs=1)

    val_dataloader = SegmentationDataModule(
        [TRAIN_SPLIT_CSV_PATH, VAL_SPLIT_CSV_PATH],
        add_adjacent_slices=True,
        batch_size=16,
        # crop_height_to_px=416,
        target_shape=(18, 320, 896),
    ).val_dataloader()

    stats = []

    random.seed(0)
    # for i, (image, _) in enumerate(val_dataloader):
    for i, (image, _) in enumerate(val_dataloader):
        try:
            print(f"\n\n==================== Processing sample {i} =====================\n")
            image = np.array(image)
            print(image.shape)

            curr_dir = directory / f"sample_{i}"
            curr_dir.mkdir(parents=True, exist_ok=True)

            cropped_height = 320
            crop_slice = get_random_crop_slice(cropped_height, image.shape[3])

            # ==================
            # Create ground truth segmentation
            # ==================

            gt_npz = onnx_seg_inference.inference(image)
            # gt = gt_npz["segmentation"][:, None, :, :]
            cropped_instance_gt = gt_npz["instances"][:, None, :, crop_slice].copy(order="C")

            # ==================
            # Create cropped input image for segmentation model (which will be used as input for instance segmentation)
            # ==================

            cropped_seg_input = image[:, :, :, crop_slice].copy(order="C")
            padding_to_896 = ((0, 0), (0, 0), (0, 0), (0, image.shape[3] - cropped_height))
            cropped_padded_seg_input = np.pad(cropped_seg_input, padding_to_896, mode="constant")
            seg_npz = onnx_seg_inference.inference(cropped_padded_seg_input)

            inst_seg_input = seg_npz["segmentation"][:, None, :, :]
            cropped_inst_seg_input = inst_seg_input[:, :, :, :height]
            padding_to_416 = ((0, 0), (0, 0), (0, 0), (0, height - cropped_height))
            cropped_input_image_with_gt = np.concatenate(
                [
                    np.pad(image[:, 1:2, :, crop_slice], padding_to_416, mode="constant"),
                    (cropped_inst_seg_input == 1),
                    (cropped_inst_seg_input == 2),
                ],
                axis=1,
            )
            cropped_input_image_with_gt = cropped_input_image_with_gt.astype(np.float32)

            # ==================
            # Create cropped input image for instance segmentation model
            # ==================

            # cropped_instance_seg_input = full_input_image_with_gt[:, :, :, crop_slice].copy(order="C")

            # if True:
            #     padding = ((0, 0), (0, 0), (0, 0), (0, image.shape[3] - height))
            #     cropped_image = np.pad(cropped_image, padding, mode="constant")
            #     cropped_gt = np.pad(cropped_gt, padding, mode="constant")

            # npz["instances"] = onnx_inst_inference.inference(cropped_image, gt=cropped_gt)["instances"]
            # npz["gt_instances"] = npz["instances"]
            new_npz = onnx_inst_inference.inference(
                cropped_input_image_with_gt, gt=np.pad(cropped_instance_gt, padding_to_416, mode="constant")
            )
            new_npz["instances"] = new_npz["instances"] * 2 - (new_npz["instances"] != 0)
            new_npz["id_to_label"] = gt_npz["id_to_label"]
            new_npz["gt_id_to_label"] = gt_npz["id_to_label"]

            new_npz["instances_post_processed"] = post_process_instances(
                new_npz["instances"], cropped_inst_seg_input[:, 0, :, :]
            )

            new_npz["cropped_segmentation"] = cropped_inst_seg_input[:, 0, :, :]

            stats.append(get_stats(new_npz["instances_post_processed"], new_npz["gt_instances"]))
            np.savez_compressed(directory / "stats.npz", stats=stats)

            plot_npz(new_npz, curr_dir / "plot.png", slices=range(0, 16))
            # if i > 50:
            #    break
            # exit(0)
        except KeyboardInterrupt:
            raise
        except:
            print(f"Error processing sample {i}")
            stats.append("Error")
            import traceback

            traceback.print_exc()
            continue

    # open_npz_and_plot_rois(new_npz, curr_dir / "plot.png")
    # open_npz_in_blender_separate_rois(npz, curr_dir / "render.png")


def main():
    eval_every_seg_sample_by_splitting()


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import Literal, Union

from spine_segmentation.cli.cli import colored_tracebacks
from spine_segmentation.datasets.sample import SampleIterator
from spine_segmentation.inference.onnx_model import ONNXInferenceModel
from spine_segmentation.plotting.plot_slice import plot_npz
from spine_segmentation.resources.paths import NAKO_DATASET_PATH
from spine_segmentation.utils.log_dir import get_next_log_dir
from spine_segmentation.visualisation.blender.open_in_blender import open_in_blender

colored_tracebacks()

import numpy as np
import onnx
from loguru import logger

from spine_segmentation.datasets.segmentation_dataset import SegmentationDataModule
from spine_segmentation.datasets.path_helper import expand_path_to_data_dirs

# onnx_model_path = "models/segmentation/2023-08-31_Seg_Unet_mit.onnx"
onnx_model_path = "models/segmentation/2023-09-12_Seg_Unet_mitb5.onnx"


def get_dataloader(train_or_val: Literal["train", "val"], path: Union[Path, str] = None, *, bs: int = 16):
    args = [] if path is None else [path]
    data_module = SegmentationDataModule(*args, augment=False, add_adjacent_slices=True, batch_size=bs, num_workers=4)
    if train_or_val == "train":
        return data_module.train_dataloader()
    elif train_or_val == "val":
        return data_module.val_dataloader()


def get_indices_from_train_dataloader(indices):
    train_data_loader = get_dataloader("train")
    samples = [sample for i, sample in enumerate(train_data_loader) if i in indices]
    return samples


def get_samples_with_l4():
    return get_indices_from_train_dataloader({13})


def get_samples_with_l6():
    return get_indices_from_train_dataloader({9, 10})


def get_sample(bs=16):
    data = get_dataloader("val", bs=bs)
    # Shape: (N, C, H, W): (16, 3, 320, 896), (16, 1, 320, 896)
    sample = next(iter(data))
    return sample


def resave_with_batch_size(onnx_path, save_to, new_batch_size=1):
    # Load the ONNX model
    model = onnx.load(onnx_path)

    # Get the name of the first input of the model
    # input_name = model.graph.input[0].name

    # Change the model's first input to the desired batch size
    model.graph.input[0].type.tensor_type.shape.dim[0].dim_value = new_batch_size
    model.graph.output[0].type.tensor_type.shape.dim[0].dim_value = new_batch_size

    # Save the modified model
    onnx.save(model, save_to)


def plot_model(onnx_path):
    import netron

    netron.start(onnx_path)


def open_npz_in_blender(numpy_path):
    spines = np.load(numpy_path)
    output, gt = spines["output"], spines["gt"]
    open_in_blender({"output": output, "gt": gt})


def open_npz_in_blender_separate_rois(npz, render_to_instead=None):
    # output, gt, instances = npz["segmentation"], npz["gt"], npz["instances"]
    # id_to_label = spines["id_to_label"].item()
    # discs = (instances % 2 == 0) * (instances != 0) * 2
    # separated_discs = separate_segmented_rois(discs) * 2
    # planes = get_planes_from_discs(separated_discs, id_to_label=id_to_label)
    # collection = BlenderCollection("planes", planes)

    open_in_blender({"instances": npz["instances"]}, render_to_instead=render_to_instead)
    # open_in_blender({"instances": instances, "gt": gt}, [collection], smooth_spine=False)


def open_npz_and_plot_rois(npz, plot_path):
    # spines = np.load(numpy_path)
    # output, gt, instances = spines["output"], spines["gt"], spines["instances"]
    # vertebrae = separate_segmented_rois(output == 1)
    # discs = separate_segmented_rois(output == 2)
    # instances = vertebrae * 2 + (discs * 2 + (discs != 0))
    plot_npz(npz, plot_path, slices=range(0, 16))


def render_all_samples(directory=None):
    if directory is None:
        directory = get_next_log_dir()
    directory = Path(directory)

    onnx_inference = ONNXInferenceModel.get_best_segmentation_model()
    # onnx_inference.load_index_list()
    # already_processed = set(onnx_inference.index_to_zip_path.values())
    already_processed = set()

    # for i, sample in enumerate(get_dataloader("train", RAW_NAKO_DATASET_PATH)):
    sample_iterator = SampleIterator(
        expand_path_to_data_dirs(NAKO_DATASET_PATH), add_adjacent_slices=True, already_processed=already_processed
    )

    try:
        # for i, (image, gt, sample, path) in enumerate(sample_iterator):
        for i, sample in enumerate(get_dataloader("train", NAKO_DATASET_PATH)):
            if i > 20:
                break
            # try:
            print(f"\n\n==================== Processing sample {i} {onnx_inference.index=} =====================\n")

            curr_dir = directory / f"sample_{i}"
            curr_dir.mkdir(parents=True, exist_ok=True)

            # sample_thing, image, gt = sample[0], sample[1]
            image, gt = sample[0].cpu().detach().numpy(), sample[1].cpu().detach().numpy()
            print(image.shape, gt.shape)

            # if sample.dicom is not None:
            # onnx_inference.index_to_series_id[onnx_inference.index] = sample.dicom.SeriesInstanceUID
            # onnx_inference.index_to_patient_id[onnx_inference.index] = sample.dicom.PatientID
            # onnx_inference.index_to_zip_path[onnx_inference.index] = sample.path

            npz = onnx_inference.inference(image, gt=gt)

            # if i % 100 == 0:
            #     onnx_inference.save_index_list()
            # except:
            #     print(f"Error processing sample {i}")
            #     continue
            npz["segmentation"] = npz["gt"].astype(int)
            # npz["instances"] = npz["gt_instances"]
            open_npz_and_plot_rois(npz, curr_dir / "plot.png")
            open_npz_in_blender_separate_rois(npz, curr_dir / "render.png")
            # open_npz_in_blender_separate_rois(npz)
    except KeyboardInterrupt:
        # print("Saving index list...")
        pass
    except:
        print("Error")
        import traceback

        traceback.print_exc()
    #    onnx_inference.save_index_list()


@logger.catch
def main():
    reuse_path = "logs/2023-09-12_154326"
    render_all_samples()


if __name__ == "__main__":
    main()



from functools import lru_cache
from pathlib import Path
from typing import Dict, Literal, Union

import numpy as np
import onnxruntime
import xxhash

from spine_segmentation.instance_separation.instance_separation import separate_rois_with_labels
from spine_segmentation.model_downloader.model_downloader import download_file_with_progress
from spine_segmentation.resources.paths import CACHE_PATH, MODELS_PATH
from spine_segmentation.utils.profiling import profile


class ONNXInferenceModel:
    def __init__(
        self,
        onnx_model_path: Union[Path, str],
        is_segmentation_model: bool,
        device: Union[Literal["CPU"], int] = "CPU",
        height: int = 896,
    ):
        self.onnx_model_path = Path(onnx_model_path)
        if not self.onnx_model_path.exists():
            raise FileNotFoundError(f"Model {self.onnx_model_path.absolute()} does not exist!")
        self.device = device
        self.cache_path = CACHE_PATH / "onnx_inference" / self.onnx_model_path.stem
        self.index_list = None
        if (self.cache_path / "index_list.npz").exists():
            self.load_index_list()

        self.height = height

        self.cache_path.mkdir(parents=True, exist_ok=True)
        self.is_segmentation_model = is_segmentation_model
        self.index = 0

        self.index_to_npz_path: Dict[int, str] = {}
        self.index_to_slice_index: Dict[int, int] = {}
        self.index_to_series_id = {}
        self.index_to_patient_id = {}
        self.index_to_zip_path = {}

    @staticmethod
    def get_best_segmentation_model(*args, **kwargs):
        model_url = "https://github.com/LiquidFun/Spine/releases/download/onnx_models_1.0.0/2023-11-06_Seg2_Unet_resnet152_896px.onnx"
        onnx_model_path = download_file_with_progress(model_url)
        return ONNXInferenceModel(onnx_model_path, *args, **kwargs, is_segmentation_model=True)

    @staticmethod
    def get_best_instance_segmentation_model(*args, **kwargs):
        model_url = "https://github.com/LiquidFun/Spine/releases/download/onnx_models_1.0.0/2023-11-09_Seg25_Unet_resnet152_vert-only_896px_no-input-seg.onnx"
        onnx_model_path = download_file_with_progress(model_url)
        return ONNXInferenceModel(onnx_model_path, *args, **kwargs, is_segmentation_model=False)

    def save_index_list(self):
        path = self.cache_path / "index_list.npz"
        np.savez_compressed(
            path,
            index_to_npz_path=self.index_to_npz_path,
            index_to_slice_index=self.index_to_slice_index,
            index_to_series_id=self.index_to_series_id,
            index_to_patient_id=self.index_to_patient_id,
            index_to_zip_path=self.index_to_zip_path,
        )

    def load_index_list(self):
        path = self.cache_path / "index_list.npz"
        index_list = np.load(path, allow_pickle=True)
        self.index = max(index_list["index_to_slice_index"].item()) + 1
        self.index_to_npz_path = index_list["index_to_npz_path"].item()
        self.index_to_slice_index = index_list["index_to_slice_index"].item()
        self.index_to_series_id = index_list["index_to_series_id"].item()
        self.index_to_patient_id = index_list["index_to_patient_id"].item()
        self.index_to_zip_path = index_list["index_to_zip_path"].item()

    @property
    @lru_cache(maxsize=1)
    def _onnx_session(self):
        device_type = "CPU" if self.device == "CPU" else "CUDA"
        # options = onnxruntime.SessionOptions()
        session_kwargs = {}
        if device_type == "CUDA":
            assert isinstance(self.device, int)
            session_kwargs["provider_options"] = [{"device_id": self.device}]
        return onnxruntime.InferenceSession(
            self.onnx_model_path, providers=[f"{device_type}ExecutionProvider"], **session_kwargs
        )

    def _hash_array(self, array: np.ndarray) -> str:
        # Make array c contiguous
        # array = np.ascontiguousarray(array)
        return xxhash.xxh128(array).hexdigest()

    def _get_cached_npz_path(self, array: np.ndarray) -> Union[None, Path]:
        hashed = self._hash_array(array)
        path = self.cache_path / "output" / f"{hashed}.npz"
        print(path)
        path.parent.mkdir(parents=True, exist_ok=True)
        return path

    @profile
    def inference(self, image: np.ndarray, *, gt: np.ndarray = None, cache: bool = True):
        """
        :param image: shape (batch_size, channels, height, width)
        :param gt: ground truth, shape (batch_size, height, width) (OPTIONAL: only used to save in cache)
        :param cache: whether to retrieve from cache if already exists
        """
        if gt is not None:
            assert gt.shape[1] == 1
            gt = gt[:, 0, :, :]
            assert len(gt.shape) == 3, f"{gt.shape=}"
            assert gt.shape == image[:, 0, :, :].shape
        assert len(image.shape) == 4, f"{image.shape=}"

        # if image.shape[3] != self.height:
        #     padding = ((0, 0), (0, 0), (0, 0), (0, self.height - image.shape[2]))
        #     image = np.pad(image, padding, mode="constant")

        cached_path = self._get_cached_npz_path(image)

        if not cached_path.exists() or not cache:
            # assert False
            input_name = self._onnx_session.get_inputs()[0].name
            onnx_inputs = {input_name: image}
            onnx_outputs = self._onnx_session.run(None, onnx_inputs)
            onnx_outputs = np.array(onnx_outputs)

            segmentation = np.argmax(onnx_outputs[0], axis=1)

            if self.is_segmentation_model:
                instances, id_to_label = separate_rois_with_labels(segmentation)
                save_arrays = dict(
                    # original=image,
                    segmentation=segmentation,
                    instances=instances,
                    id_to_label=id_to_label,
                )

                assert (
                    image[:, 0, :, :].shape == segmentation.shape == instances.shape
                ), f"{segmentation.shape=} != {instances.shape=} != {image[:, 0, :, :].shape=}"

            else:
                save_arrays = dict(instances=segmentation)

            np.savez_compressed(cached_path, **save_arrays)

        npz = np.load(cached_path, allow_pickle=True)

        self.index_to_npz_path[self.index] = cached_path
        for i in range(len(npz["instances"])):
            self.index_to_slice_index[self.index] = i
            self.index += 1

        print(f"{len(self.index_to_npz_path)=} {len(self.index_to_slice_index)=}")

        npz_dict = dict(npz)
        npz_dict["original"] = image
        if gt is not None:
            key = "gt" if self.is_segmentation_model else "gt_instances"
            npz_dict[key] = gt
        return npz_dict



import numpy as np
import torch
from torch_geometric.data import Data, InMemoryDataset

from spine_segmentation.resources.preloaded import get_measure_statistics


class VectorTableGraphs(InMemoryDataset):
    def __init__(self, root, transform=None, pre_transform=None, pre_filter=None, device="cpu"):
        super().__init__(root, transform, pre_transform, pre_filter)
        self.data, self.slices = torch.load(self.processed_paths[0], map_location=device)

    @property
    def raw_file_names(self):
        return [""]

    @property
    def processed_file_names(self):
        return ["graphs_all.pt"]

    def download(self):
        pass

    def get_graphs(self, vector_table, size=10):
        patients, rois, axes = vector_table.shape
        vector_table = np.nan_to_num(vector_table)
        num_classes = diff = rois - size
        graphs = []
        r = np.arange(size - 1)
        edges = torch.tensor([r, r + 1], dtype=torch.long)
        for g in range(patients):
            # If mod 100
            if g % 100 == 0:
                print(g)
            for start in range(diff):
                features = vector_table[g, start : start + size, :]
                label = torch.tensor([start])
                graph = Data(x=features, edge_index=edges, y=label)
                graphs.append(graph)
        return graphs

    def process(self):
        # Read data into huge `Data` list.
        vector_table = get_measure_statistics().filter_by_column(
            excluded_substrings=["_PROB", "_DIFF", "ANGLE_COR", "ANGLE_SAG", "HEIGHT_MIN"],
            reshape=True,
        )
        data_list = self.get_graphs(vector_table)

        if self.pre_filter is not None:
            data_list = [data for data in data_list if self.pre_filter(data)]

        if self.pre_transform is not None:
            data_list = [self.pre_transform(data) for data in data_list]

        data, slices = self.collate(data_list)
        data.x = torch.Tensor(data.x)
        torch.save((data, slices), self.processed_paths[0])



import lightning.pytorch as pl
import numpy as np
from torch.utils.data import DataLoader, Dataset

from spine_segmentation.resources.preloaded import get_measure_statistics


class FeatureDataset(Dataset):
    def __init__(self, vector_table, size=10):
        patients, rois, axes = vector_table.shape
        # vector_table = np.nan_to_num(vector_table)
        self.num_classes = diff = rois - size
        samples = patients * diff
        vector_table = np.nan_to_num(vector_table)

        self.data = np.zeros((samples, size, axes)).astype(np.float32)
        self.labels = np.zeros((samples,)).astype(int)
        for start in range(0, diff):
            sample = vector_table[:, start : start + size, :]
            # print(np.isnan(sample).sum())
            # if not np.isnan(sample).any():
            self.data[start::diff, :, :] = sample
            self.labels[start::diff] = start + 1

        # self.data = np.zeros((samples, size, axes)).astype(np.float32)
        # self.labels = np.zeros((samples,)).astype(int)
        # start_indices = np.arange(diff)
        # self.data[start_indices[:, None] * diff + np.arange(samples // diff)[:, None, None], :, :] \
        #     = vector_table[:, start_indices[:, None] + np.arange(size)[None, :, None], :]
        # self.labels[start_indices[:, None] * diff + np.arange(samples // diff)[:, None]] = start_indices[:, None]

        # shape: (10425*37, 10, 3)
        # self.data = np.zeros((samples, size, axes)).astype(np.float32)
        # self.labels = np.zeros((samples,)).astype(int)

        # # create an array of start indices [[0], [1], ..., [diff-1]], shape: (diff, 1)
        # start_indices = np.arange(diff)[:, None]

        # # Compute the indices in self.data and self.labels where the values will be assigned.
        # into_indices = start_indices * diff + np.arange(samples // diff)[:, None, None]
        # from_indices = start_indices + np.arange(size)[None, :, None]

        # # Assign the values from vector_table to self.data
        # self.data[into_indices, :, :] = vector_table[:, from_indices, :]

        # # Assign the values from start_indices to self.labels
        # self.labels[into_indices.squeeze()] = start_indices

        # assert np.all(self.data_slow == self.data)
        # assert np.all(self.labels_slow == self.labels)

    def __getitem__(self, index):
        x = self.data[index]
        y = self.labels[index]
        return x, y

    def __len__(self):
        return self.data.shape[0]


class FeatureDataModule(pl.LightningDataModule):
    def __init__(
        self, data_dir=None, batch_size: int = 1, num_workers: int = 2, sample_size=10, train_split: float = 0.9
    ):
        super().__init__()
        stats = get_measure_statistics()
        only_for_vertebra = ["ANGLE_COR", "ANGLE_SAG", "HEIGHT_MIN"]
        too_correlated = ["_CENTER"]
        exclude = only_for_vertebra + too_correlated
        # self.features = stats.filter_by_column(
        #     required_substrings=["DIRECTION"], excluded_substrings=["CANAL_DIRECTION"], reshape=True
        # )

        self.features = stats.filter_by_column(
            required_substrings=["DIRECTION", "HEIGHT_MAX", "VOLUME", "LENGTH_AP"],
            excluded_substrings=["CANAL_DIRECTION", "VOLUME_PROB", "VOLUME_DIFF"],
            reshape=True,
        )
        # self.features = stats.filter_by_column(excluded_substrings=exclude, reshape=True)
        self.batch_size = batch_size
        train_size = round(self.features.shape[0] * train_split)
        train_features = self.features[:train_size]
        val_features = self.features[train_size:]

        self.dataloader_kwargs = dict(
            batch_size=batch_size,
            num_workers=num_workers,
        )

        dataset_kwargs = dict(
            size=sample_size,
        )

        self.train_dataset = FeatureDataset(train_features, **dataset_kwargs)
        self.val_dataset = FeatureDataset(val_features, **dataset_kwargs)

    def train_dataloader(self):
        return DataLoader(self.train_dataset, shuffle=True, **self.dataloader_kwargs)

    def val_dataloader(self):
        return DataLoader(self.val_dataset, **self.dataloader_kwargs)


if __name__ == "__main__":
    stats = get_measure_statistics()
    FeatureDataset(stats.canal_direction_vectors)



import mlflow
from lightning.pytorch.callbacks import LearningRateFinder

from spine_segmentation.utils.log_dir import get_next_log_dir


class FineTuneLearningRateFinder(LearningRateFinder):
    def __init__(self, milestones, *args, update_attr=True, **kwargs):
        super().__init__(*args, update_attr=update_attr, early_stop_threshold=None, num_training_steps=300, **kwargs)
        self.milestones = milestones

    def on_fit_start(self, *args, **kwargs):
        return

    def on_train_epoch_start(self, trainer, pl_module):
        if trainer.current_epoch in self.milestones or trainer.current_epoch == 0:
            lr_before = trainer.optimizers[0].param_groups[0]["lr"]
            self.lr_find(trainer, pl_module)
            trainer.optimizers[0].param_groups[0]["lr"] = self.optimal_lr.suggestion()
            fig = self.optimal_lr.plot(suggest=True)
            fig_path = get_next_log_dir() / f"lr_finder{trainer.current_epoch:03}.png"
            fig.savefig(fig_path)
            mlflow.log_artifact(fig_path)
            lr_after = trainer.optimizers[0].param_groups[0]["lr"]
            print(f"Optimizer learning rate before {lr_before}, and after: {lr_after}")



import re
from collections import OrderedDict
from pathlib import Path
from typing import Dict

import numpy as np
import pandas as pd

from spine_segmentation.resources.paths import DATASETS_PATH, NAKO_DATASET_PATH

SPLITS_PATH = DATASETS_PATH / "nako_splits"
SPLITS_PATH.mkdir(parents=True, exist_ok=True)
TRAIN_SEG_PATH = SPLITS_PATH / "train_seg.csv"
VAL_SEG_PATH = SPLITS_PATH / "val_seg.csv"
TRAIN_INSTSEG_PATH = SPLITS_PATH / "train_instseg.csv"
VAL_INSTSEG_PATH = SPLITS_PATH / "val_instseg.csv"


TEST_PATIENT_ID_FILE_PATH = Path(__file__).parent / "test_patient_ids.txt"

INDEX_LIST_PATH = Path.home() / "<define-this-path>"


def save_as_csv(pat_id_to_path: Dict, path):
    print(f"Saving {len(pat_id_to_path)} to {path}")
    pat_id_to_path = OrderedDict(sorted(pat_id_to_path.items()))
    df = pd.DataFrame.from_dict(pat_id_to_path, orient="index", columns=["Path"])
    df["Path"] = df["Path"].astype(str)
    prefixes = [
        "",
    ]
    for prefix in prefixes:
        assert prefix[-1] != "/", f"Prefix {prefix} should not end with '/'"
        df["Path"] = df["Path"].str.replace(prefix, "~", regex=False)
    df.to_csv(path)


def main():
    full_dataset = {}
    full_gt_dataset = {}
    gt_only = {}

    test_set_ids = TEST_PATIENT_ID_FILE_PATH.open().read().strip().split("\n")
    print(test_set_ids)

    trainval_raw_dataset = {}
    trainval_truthed_dataset = {}

    test_set = {}

    # for path in RAW_NAKO_DATASET_PATH.glob("*"):
    #     pat_id = path.name.replace("_30_Sag T2 Spine.zip", "")
    #     full_dataset[pat_id] = path

    #     if pat_id not in test_set_ids:
    #         trainval_raw_dataset[pat_id] = path

    npz = np.load(INDEX_LIST_PATH, allow_pickle=True)

    for index, path in npz["index_to_zip_path"].item().items():
        pat_id = re.sub(r"_\d[05]_sag t2 spine.zip", "", path.name.lower())
        assert re.fullmatch(r"1\d{5}", pat_id), f"Invalid patient id {pat_id}"
        full_dataset[pat_id] = path

        if pat_id not in test_set_ids:
            trainval_raw_dataset[pat_id] = path

    for path in NAKO_DATASET_PATH.glob("*"):

        def get_gt_type(p):
            return p.name.replace(".", "_").split("_")[1]

        gt_types = [get_gt_type(p) for p in path.glob("*_*.nii.gz")]
        pat_id = path.name
        if "WK" in gt_types and "BS" in gt_types:
            if path.name not in full_dataset:
                gt_only[pat_id] = path
            full_gt_dataset[pat_id] = path
            full_dataset[pat_id] = path

            if pat_id not in test_set_ids:
                trainval_truthed_dataset[pat_id] = path
            else:
                test_set[pat_id] = path
        else:
            if pat_id in test_set_ids:
                print("testset patid no WK or BS", pat_id)

    assert all(pid in full_gt_dataset for pid in test_set_ids), "Not all test set ids not in gt_dataset"

    # paths.append(path.name)
    print(len(gt_only))
    print(*gt_only)

    seg_train_set = trainval_truthed_dataset
    seg_val_set = test_set

    import random

    random.seed(42)
    train_percentage = 0.95
    train_size = int(len(trainval_raw_dataset) * train_percentage)
    trainval_pat_ids = list(trainval_raw_dataset.keys())
    random.shuffle(trainval_pat_ids)
    train_ids, val_ids = trainval_pat_ids[:train_size], trainval_pat_ids[train_size:]

    union = set(seg_train_set) & set(val_ids)
    val_ids = set(val_ids) - union
    train_ids = set(train_ids) | union

    instseg_train_set = {pid: trainval_raw_dataset[pid] for pid in train_ids}
    instseg_val_set = {pid: trainval_raw_dataset[pid] for pid in val_ids}
    print(len(instseg_train_set), len(instseg_val_set))
    print(list(instseg_val_set)[:10])

    assert seg_val_set.keys() & instseg_train_set.keys() == set(), "Train and val sets should be disjoint"
    assert seg_val_set.keys() & instseg_val_set.keys() == set(), "Train and val sets should be disjoint"
    assert seg_val_set.keys() & seg_train_set.keys() == set(), "Train and val sets should be disjoint"

    assert instseg_val_set.keys() & instseg_train_set.keys() == set(), "Train and val sets should be disjoint"
    assert (
        instseg_val_set.keys() & seg_train_set.keys() == set()
    ), f"Should be disjoint, but got {instseg_val_set.keys() & seg_train_set.keys()}"

    save_as_csv(seg_train_set, TRAIN_SEG_PATH)
    save_as_csv(seg_val_set, VAL_SEG_PATH)
    save_as_csv(instseg_train_set, TRAIN_INSTSEG_PATH)
    save_as_csv(instseg_val_set, VAL_INSTSEG_PATH)


if __name__ == "__main__":
    main()



from typing import Dict, List, Tuple

import cc3d
import numpy as np

from spine_segmentation.dataloader.statistics import Statistics
from spine_segmentation.spine_types.labels import get_labels_for_n_classes
from spine_segmentation.spine_types.line_segment import LineSegment
from spine_segmentation.spine_types.plane import Plane


def mask_rare_rois(arr3d, min_volume: int) -> np.ndarray:
    arr3d = arr3d.copy()
    for i, count in zip(*np.unique(arr3d, return_counts=True)):
        if count < min_volume:
            arr3d[arr3d == i] = 0
    return arr3d


def separate_segmented_rois(
    arr3d: np.ndarray, min_volume=100
) -> np.ndarray:  # if any(name in str(path_out) for name in blacklist_num):
    # return False
    # arr3d[arr3d != 0] = 1

    # print(np.unique(arr3d,return_counts=True))
    seg, N = cc3d.connected_components(arr3d, connectivity=26, return_N=True)

    i_ = 1
    for i in range(1, N + 1):
        arr3d = seg == i_
        size = np.count_nonzero(arr3d)
        if size < min_volume:
            # skip this!
            # make to 0
            seg = np.where(arr3d, 0, seg)
            # shift >i down
            seg = np.where(seg > i_, seg - 1, seg)
            continue
        i_ += 1
    i_ -= 1
    # if i_ != 22:
    # print(color_error(f"found {i_} components from originally {N}."))
    # raise ValueError(f"Wrong number of components {i_}")
    # return False
    # order numbers from top to down
    N = i_

    def to_struct(i: int) -> Tuple:
        bool_map = seg == i
        pos = -np.average(np.where(bool_map)[2])
        num = np.sum(bool_map)
        return pos, num, bool_map, i

    ypos_map_idx = [to_struct(i) for i in range(1, N + 1)]
    if sorted(ypos_map_idx) != ypos_map_idx:
        # reorder indexes
        lst = sorted(ypos_map_idx)
        pos, num, bool_map, i = lst[0]
        merged_map = bool_map.astype(int)
        for idx, (pos_, num_, bool_map_, i_) in enumerate(lst[1:], start=2):
            merged_map = np.where(bool_map_, idx, merged_map)
        seg = merged_map
    return seg


def separate_segmented_rois_by_planes(arr3d: np.ndarray, planes: List[Plane], use_plane_index):
    seg = np.array(arr3d)
    coords = np.array(np.nonzero(seg)).T
    assert coords.shape[1] == 3

    # Set any point above, if there are any, as it does not have to be that all the point are below
    if use_plane_index:
        first_plane = planes[0]
        coords_above = coords[first_plane.are_points_above_plane(coords)].T
        seg[coords_above[0], coords_above[1], coords_above[2]] = first_plane.index - 1

    for plane in planes:
        coords_below = coords[~plane.are_points_above_plane(coords)].T
        # print(f"{coords_below.shape=}, {coords.shape=} {seg.shape=}")
        assert coords_below.shape[0] == 3
        if use_plane_index:
            seg[coords_below[0], coords_below[1], coords_below[2]] = plane.index + 1
        else:
            seg[coords_below[0], coords_below[1], coords_below[2]] += 1
    return seg


def __get_planes_from_discs_with_eigenvectors(separated_discs: np.ndarray, id_to_label=None):
    planes = []
    for i in np.unique(separated_discs)[1:]:
        coords = np.array(np.nonzero(separated_discs == i)).T
        centroid = np.mean(coords, axis=0)
        deviations = coords - centroid
        covariance_matrix = np.cov(deviations, rowvar=False)
        eigenvalues, eigenvectors = np.linalg.eigh(covariance_matrix)
        normal_vector = eigenvectors[:, -1]
        normal_vector = normal_vector / np.linalg.norm(normal_vector)
        name = None if id_to_label is None else id_to_label[i]

        planes.append(Plane(centroid, normal_vector, name=name))
        for j in range(eigenvectors.shape[1]):
            points = [centroid, centroid + eigenvectors[:, j] * 10]
            planes.append(LineSegment(points, name=name + f"_eigen-{j + 1}"))
        # planes.append(LineSegment([centroid, centroid + normal_vector * 10], name=name + "normal-vec"))

    return planes


def _infer_additional_center_points_at_top_and_bottom(centroids: np.ndarray):
    centroids_as_numpy = np.array(centroids)[None, :, :]
    first4 = centroids_as_numpy[:, :4, :]
    last4 = centroids_as_numpy[:, -4:, :][
        :, ::-1, :
    ]  # Needs to be done in 2 steps, as otherwise -4::-1 has a different meaning
    first_predicted = Statistics.predict_next_center_point(first4)[:, -1, :]
    last_predicted = Statistics.predict_next_center_point(last4)[:, -1, :]

    return np.concatenate([first_predicted, centroids_as_numpy[0], last_predicted], axis=0)


def get_planes_from_rois(separated_rois: np.ndarray, id_to_label=None):
    planes = []
    centroids = []
    labels = []
    indices = []
    for i in np.unique(separated_rois)[1:]:
        coords = np.array(np.nonzero(separated_rois == i)).T
        centroid = np.mean(coords, axis=0)
        centroids.append(centroid)
        labels.append(str(i) if id_to_label is None else id_to_label[i])
        indices.append(i)

    if len(centroids) == 0:
        return []

    expanded = _infer_additional_center_points_at_top_and_bottom(centroids)

    # planes.append(LineSegment(expanded, name="spine_as_line"))

    for label, index, lower, current, upper in zip(labels, indices, expanded[2:], expanded[1:-1], expanded[:-2]):
        normal = upper - lower
        normal = normal / np.linalg.norm(normal)

        planes.append(Plane(current, normal, name=label, index=index))
        # planes.append(LineSegment([current, current+normal * 10], name=label + f"_normal"))

    return planes


def separate_rois_with_labels(arr3d: np.ndarray, *, split_along_discs=True) -> Tuple[np.ndarray, Dict[int, str]]:
    arr3d = arr3d.copy()
    assert arr3d.ndim == 3, f"Expected 3d array, got {arr3d.shape}"
    assert list(np.unique(arr3d)) == [
        0,
        1,
        2,
    ], f"Expected arr3d to be only 'background, vertebrae, discs', got: {np.unique(arr3d)}"

    groups = cc3d.connected_components(arr3d != 0, connectivity=26)
    uniq = np.unique(groups, return_counts=True)

    # Find the largest connected component in the image, this filters any small objects which are not connected
    # to the main spine object!
    # Take -2 because the most common will be the background
    most_common_id = sorted(zip(*uniq), key=lambda x: x[1])[-2][0]
    arr3d[groups != most_common_id] = 0

    if split_along_discs:
        discs = separate_segmented_rois(arr3d == 2)
        planes = get_planes_from_rois(discs)
        # vertebrae = separate_segmented_rois(arr3d == 1)
        vertebrae = separate_segmented_rois_by_planes((arr3d == 1).astype(int), planes, use_plane_index=False)
    else:
        discs = separate_segmented_rois(arr3d == 2)
        vertebrae = separate_segmented_rois(arr3d == 1)

    # assert len(np.unique(vertebrae)) == len(VERTEBRA_LABELS) + 1, f"Expected {len(VERTEBRA_LABELS) + 1} vertebrae, got {len(np.unique(vertebrae))}"
    # assert len(np.unique(discs)) == len(DISC_LABELS) + 1, f"Expected {len(DISC_LABELS) + 1} discs, got {len(np.unique(discs))}"

    instances = (vertebrae * 2) - (vertebrae != 0) + (discs * 2)
    # print(list(zip(*np.unique(instances, return_counts=True))))

    ids = np.unique(instances)[1:]
    labels = get_labels_for_n_classes(49)
    id_to_label = dict(zip(range(1, 50), labels))
    id_to_label = {id: label for id, label in id_to_label.items() if id in ids}
    return instances, id_to_label

    from matplotlib import pyplot as plt

    plt.imshow(discs[8])
    plt.show()



from functools import partial
from multiprocessing import Manager, Pool, cpu_count

import numpy as np
from tqdm import tqdm

from spine_segmentation.cli.cli import colored_tracebacks
from spine_segmentation.inference.onnx_model import ONNXInferenceModel

colored_tracebacks()

common_cases = [
    "C2:C2-C3:C3:C3-C4:C4:C4-C5:C5:C5-C6:C6:C6-C7:C7:C7-T1:T1:T1-T2:T2:T2-T3:T3:T3-T4:T4:T4-T5:T5:T5-T6:T6:T6-T7:T7"
    ":T7-T8:T8:T8-T9:T9:T9-T10:T10:T10-T11:T11:T11-T12:T12:T12-L1:L1:L1-L2:L2:L2-L3:L3:L3-L4:L4:L4-L5:L5:L5-S1:S1",
    "",
    "C2:C2-C3:C3:C3-C4:C4:C4-C5:C5:C5-C6:C6:C6-C7:C7:C7-T1:T1:T1-T2:T2:T2-T3:T3:T3-T4:T4:T4-T5:T5:T5-T6:T6:T6-T7:T7"
    ":T7-T8:T8:T8-T9:T9:T9-T10:T10:T10-T11:T11:T11-T12:T12:T12-L1:L1:L1-L2:L2:L2-L3:L3:L3-L4:L4:L4-L5:L5:L5-L6:L6:L6"
    "-S1:S1",
    "",
    "C2:C2-C3:C3:C3-C4:C4:C4-C5:C5:C5-C6:C6:C6-C7:C7:C7-T1:T1:T1-T2:T2:T2-T3:T3:T3-T4:T4:T4-T5:T5:T5-T6:T6:T6-T7:T7"
    ":T7-T8:T8:T8-T9:T9:T9-T10:T10:T10-T11:T11:T11-T12:T12:T12-L1:L1:L1-L2:L2:L2-L3:L3:L3-L4:L4:L4-S1:S1",
]


def process_path(path, counter, lock, rare_cases):
    npz = np.load(path, allow_pickle=True)
    id_to_label = npz["id_to_label"].item()
    key = ":".join(id_to_label.values())
    with lock:
        if key not in counter:
            counter[key] = 0
        counter[key] += 1
        if key not in common_cases:
            print(f"Rare case: {path}")
            rare_cases.append(path)


def main():
    onnx = ONNXInferenceModel.get_best_segmentation_model()
    onnx.load_index_list()
    # Tqdm loading and multiprocessing:
    # path = (Path.home() / "devel/src/git" / path.relative_to(relative_to)).expanduser()

    # Assuming onnx.index_to_npz_path.values() is a list of paths
    paths = list(onnx.index_to_npz_path.values())

    with Manager() as manager:
        counter = manager.dict()  # shared counter among processes
        rare_cases = manager.list()
        counter_lock = manager.Lock()  # lock to safely update counter

        func = partial(process_path, counter=counter, lock=counter_lock, rare_cases=rare_cases)

        with Pool(cpu_count()) as pool:
            for _ in tqdm(pool.imap_unordered(func, paths), total=len(paths)):
                pass

        # Convert manager counter to a normal dictionary
        counter_dict = dict(counter)
        rare_cases = list(rare_cases)

    for key, value in counter_dict.items():
        print(f"{value}: len={len(key)} {key}")

    for case in rare_cases:
        print(case)


if __name__ == "__main__":
    main()
